#if !defined(__AABB_H)
#define __AABB_H

/* This file contains implementations of 2D and 3D axis-aligned boxes */

struct AABB {
public:
	Point3 min;
	Point3 max;

	/// Construct an invalid bounding box
	inline AABB() {
		reset();
	}

	/// Create a bounding box from two 3-dimension vectors
	inline AABB(const Point3 &min, const Point3 &max)
		: min(min), max(max) {
	}

	/// Copy-constructor
	inline AABB(const AABB &aabb) 
		: min(aabb.min), max(aabb.max) {
	}

	inline Point3 getCorner(int corner) const {
		return Point3(corner & 1 ? max.x : min.x,
				corner & 2 ? max.y : min.y,
				corner & 4 ? max.z : min.z);
	}

	inline Point3 getMidPoint3() const {
		return Point3(
			min.x + max.x,
			min.y + max.y,
			min.z + max.z
		) * 0.5f;
	}

	inline bool contains(const Point3 &vec) const {
		return isValid() &&
			(vec.x >= min.x && vec.x <= max.x) &&
			(vec.y >= min.y && vec.y <= max.y) &&
			(vec.z >= min.z && vec.z <= max.z);
	}

	inline void expandBy(const Point3 &vec) {
		min.x = std::min(min.x, vec.x);
		min.y = std::min(min.y, vec.y);
		min.z = std::min(min.z, vec.z);
		max.x = std::max(max.x, vec.x);
		max.y = std::max(max.y, vec.y);
		max.z = std::max(max.z, vec.z);
	}

	inline void expandBy(const Point3s &vec) {
		min.x = std::min(min.x, (Float) vec.x);
		min.y = std::min(min.y, (Float) vec.y);
		min.z = std::min(min.z, (Float) vec.z);
		max.x = std::max(max.x, (Float) vec.x);
		max.y = std::max(max.y, (Float) vec.y);
		max.z = std::max(max.z, (Float) vec.z);
	}

	inline void expandBy(const AABB &aabb) {
		min.x = std::min(min.x, aabb.min.x);
		min.y = std::min(min.y, aabb.min.y);
		min.z = std::min(min.z, aabb.min.z);
		max.x = std::max(max.x, aabb.max.x);
		max.y = std::max(max.y, aabb.max.y);
		max.z = std::max(max.z, aabb.max.z);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "AABB[";
		if (!isValid()) {
			oss << "invalid";
		} else {
			oss << "min=" << min.toString()
				<< ", max=" << max.toString();
		}
		oss	<< "]";
		return oss.str();
	}

	/// Assignment operator
	inline AABB &operator=(const AABB &aabb) {
		min = aabb.min;
		max = aabb.max;
		return *this;
	}

	/// Mark the bounding box as invalid
	inline void reset() {
		const Float inf = std::numeric_limits<Float>::infinity();
		min = Point3(inf, inf, inf);
		max = Point3(-inf, -inf, -inf);
	}

	/// Calculate the volume of the bounding box
	inline Float getVolume() const {
		Float x = max.x - min.x;
		Float y = max.y - min.y;
		Float z = max.z - min.z;

		return x*x + y*y + z*z;
	}

	/// Clip to another AABB
	inline void clip(const AABB &aabb) {
		min.x = std::max(min.x, aabb.min.x);
		min.y = std::max(min.y, aabb.min.y);
		min.z = std::max(min.z, aabb.min.z);
		max.x = std::min(max.x, aabb.max.x);
		max.y = std::min(max.y, aabb.max.y);
		max.z = std::min(max.z, aabb.max.z);
	}

	/// Return the center point
	inline Point3 getCenter() const {
		return Point3(max + min) / 2.0f;
	}
	
	/// Return the extents
	inline Vector3 getExtents() const {
		return max - min;
	}

	/// Calculate the surface area of the bounding box
	inline Float getSurfaceArea() const {
		Vector3 d = max - min;
		return 2.0f * (d.x*d.y + d.x*d.z + d.y*d.z);
	}

	/// Return the axis with the largest corresponding AABB side
	inline int getLargestAxis() const {
		Vector3 d = max - min;
		if (d.x > d.y && d.x > d.z)
			return 0;
		else if (d.y > d.z)
			return 1;
		else
			return 2;
	}

	/// Return whether this bounding box is valid
	inline bool isValid() const {
		return max.x >= min.x && max.y >= min.y && max.z >= min.z;
	}

	/**
	 * Return whether this bounding box covers a non-zero
	 * amount of space
	 */
	inline bool isEmpty() const {
		return max.x <= min.x
			&& max.y <= min.y
			&& max.z <= min.z;
	}

	/** \brief Calculate the near and far ray-AABB intersection
	 * points (if they exist).
	 */
	inline bool rayIntersect(const Ray3 &ray, Float &nearT, Float &farT) const {
		nearT = -std::numeric_limits<Float>::infinity();
		farT  = std::numeric_limits<Float>::infinity();

		/* For each pair of AABB planes */
		for (int i=0; i<3; i++) {
			const Float direction = ray.d[i];
			const Float origin = ray.o[i];
			const Float minVal = min[i], maxVal = max[i];

			if (direction == 0) {
				/* The ray is parallel to the planes */
				if (origin < minVal || origin > maxVal)
					return false;
			} else {
				/* Calculate intersection distances */
				Float t1 = (minVal - origin) / ray.d[i];
				Float t2 = (maxVal - origin) / ray.d[i];

				if (t1 > t2) {
					Float tmp = t1;
					t1 = t2;
					t2 = tmp;
				}

				nearT = std::max(nearT, t1);
				farT = std::min(farT, t2);

				if (nearT > farT)
					return false;
			}
		}
		return true;
	}


	/// Return the minimum vector of the bounding box
	inline const Point3 &getMinimum() const {
		return min;
	}

	/// Return the maximum vector of the bounding box
	inline const Point3 &getMaximum() const {
		return max;
	}
};

struct AABB2 {
public:
	Point2 min;
	Point2 max;

	/// Construct an invalid bounding box
	inline AABB2() {
		reset();
	}

	/// Create a bounding box from two 3-dimension vectors
	inline AABB2(const Point2 &min, const Point2 &max)
		: min(min), max(max) {
	}

	/// Copy-constructor
	inline AABB2(const AABB2 &aabb) 
		: min(aabb.min), max(aabb.max) {
	}

	inline bool contains(const Point2 &vec) const {
		return isValid() &&
			(vec.x >= min.x && vec.x <= max.x) &&
			(vec.y >= min.y && vec.y <= max.y);
	}

	inline void expandBy(const Point2 &vec) {
		min.x = std::min(min.x, vec.x);
		min.y = std::min(min.y, vec.y);
		max.x = std::max(max.x, vec.x);
		max.y = std::max(max.y, vec.y);
	}

	inline void expandBy(const AABB2 &aabb) {
		min.x = std::min(min.x, aabb.min.x);
		min.y = std::min(min.y, aabb.min.y);
		max.x = std::max(max.x, aabb.max.x);
		max.y = std::max(max.y, aabb.max.y);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "AABB2[";
		if (!isValid()) {
			oss << "invalid";
		} else {
			oss << "min=" << min.toString()
				<< ", max=" << max.toString();
		}
		oss	<< "]";
		return oss.str();
	}

	/// Assignment operator
	inline AABB2 &operator=(const AABB2 &aabb) {
		min = aabb.min;
		max = aabb.max;
		return *this;
	}

	/// Mark the bounding box as invalid
	inline void reset() {
		const Float inf = std::numeric_limits<Float>::infinity();
		min = Point2(inf, inf);
		max = Point2(-inf, -inf);
	}

	/// Calculate the volume of the bounding box
	inline Float getVolume() const {
		Float x = max.x - min.x;
		Float y = max.y - min.y;

		return x*x + y*y;
	}

	/// Return the center point
	inline Point2 getCenter() const {
		return Point2(max + min) / 2.0f;
	}
	
	/// Return the extents
	inline Vector2 getExtents() const {
		return max - min;
	}

	/// Return whether this bounding box is valid
	inline bool isValid() const {
		return max.x >= min.x && max.y >= min.y;
	}

	/**
	 * Return whether this bounding box covers a non-zero
	 * amount of space
	 */
	inline bool isEmpty() const {
		return max.x <= min.x
			&& max.y <= min.y;
	}


	/// Return the minimum vector of the bounding box
	inline const Point2 &getMinimum() const {
		return min;
	}

	/// Return the maximum vector of the bounding box
	inline const Point2 &getMaximum() const {
		return max;
	}
};

#endif
