#if !defined(__DATAMODEL_H)
#define __DATAMODEL_H

#include <hair/parameters.h>
#include <hair/kdtree.h>
#include <hair/imageio.h>
#include <hair/timer.h>
#include <QtGui>

#define GROW_STEPSIZE 5      // in pixels
#define MAX_BLIND_STEPS 10
#define MAX_FIT_ERROR .15

/* Stores the vertices of a single hair fiber */
struct HairFiber {
	std::vector<Point3> vertices;

	// From which angles was this hair fiber seen?
	bool seen[MAX_ACTIVESET_SIZE];

	// Saturation for visualization purposes
	Float saturation;

	inline HairFiber() {
		for (int i=0; i<MAX_ACTIVESET_SIZE; ++i)
			seen[i] = false;
		saturation = (rand() / (float) RAND_MAX)  * .7f + .3f;
	}
};

/**
 * This data structure encapsulates information about one end of a 
 * partially grown hair fiber, as well as observed data that can 
 * be used to further extend it into this direction.
 */
struct ActiveSet {
	/* Currend endpoint */
	Point3 pOriginal, p;

	/* Direction at the current endpoint */
	Vector3 dOriginal, d;

	/* Observed 2D orientations for each observation angle */
	Float orientation[MAX_ACTIVESET_SIZE];

	/* Currently active observation angles */
	bool active[MAX_ACTIVESET_SIZE];

	/* Number of true entries of the previous array */
	int nActive;

	/* Partially grown hair fiber */
	HairFiber fiber;

	/* When a fiber can't be extended, a certain number of blind
	   steps are taken. This tracks the number of such steps */
	int blindSteps;

	/* Total number of fiber-growing steps so far */
	int nSteps;

	/* Set of hair observations used during the fitting process 
	   (these will be removed after the fiber has been grown)
	*/
	std::set<KDTree::Observation *> observations[MAX_ACTIVESET_SIZE];

	/* Protects the fiber data structure */
	mutable pthread_mutex_t mutex;

	inline ActiveSet() {
		for (int i=0; i<MAX_ACTIVESET_SIZE; ++i)
			active[i] = false;
		nActive = blindSteps = nSteps = 0;
		pthread_mutex_init(&mutex, NULL);
	}

	inline void lock() const {
		pthread_mutex_lock(&mutex);
	}

	inline void unlock() const {
		pthread_mutex_unlock(&mutex);
	}

	inline void operator=(const ActiveSet &a) {
		lock();
		pOriginal = a.pOriginal;
		p = a.p;
		dOriginal = a.dOriginal;
		d = a.d;
		nActive = a.nActive;
		fiber = a.fiber;
		blindSteps = a.blindSteps;
		nSteps = a.nSteps;
		for (int i=0; i<MAX_ACTIVESET_SIZE; ++i) {
			orientation[i] = a.orientation[i];
			active[i] = a.active[i];
			observations[i] = a.observations[i];
		}
		unlock();
	}

	inline ~ActiveSet() {
		pthread_mutex_destroy(&mutex);
	}
};

class WorkerThread;
class GLWidget;

/** Data model of the fiber-growing application. Stores all
 * relevant content and sends update events to registered listeners
 */
class DataModel : public QObject {
	Q_OBJECT
public:
	enum EChangeLevel {
		Repaint = 0,
		FocusAndRefresh = 1,
		Refresh = 2,
		NewFile = 3
	};

	DataModel();
	~DataModel();

	/* Oriented bounding box query (only spatially) */
	void obbQuery(const OBB &obb, int angle, std::vector<KDTree::Observation *> &result) const;

	/* Oriented bounding box query (permits specification of a 2D orientation search region) */
	void obbQuery(const OBB &obb, Float orientation, Float delta, int angle, 
		std::vector<KDTree::Observation *> &result) const;

	/* ==================================================================== */
	//    Fiber-growing-related methods

	/* Acquire updated 2D orientations after a step has been taken */
	void findOrientations(ActiveSet &activeSet) const;

	/* SVD-based direction fitting */
	bool solveDirection(ActiveSet &activeSet, Vector3 &direction) const;

	/* Re-centering correction */
	bool solveCorrection(ActiveSet &activeSet, const Point3 &proposed, Vector3 &correction) const;

	/* Perform one growing step */
	bool grow(ActiveSet &activeSet);

	void go();

	/* ==================================================================== */
	//    Content-related methods

	/// Open a new dataset
	void load(const Parameters &parameters);
	void clearFibers();
	void addFiber(const HairFiber &fiber);
	void saveOBJ(const std::string &filename);
	void saveMDL(const std::string &filename, const std::vector<HairFiber> &fibers) const;
	void loadMDL(const std::string &filename, bool comparison);
	inline const std::vector<HairFiber> &getHairFibers() const { return m_fibers; }
	inline const std::vector<HairFiber> &getComparisonHairFibers() const { return m_comparisonFibers; }
	

	/* ==================================================================== */
	//    View-related methods

	inline const AABB &getAABB() const { return m_aabb; }
	inline void setUpdateViews(bool value) { m_updateViews = value; }
	inline bool getUpdateViews() const { return m_updateViews; }
	const ActiveSet &getActiveSet(int i) const;
	inline int getActiveSetCount() const { return m_threads.size() * 2; }
	inline int getThreadCount() const { return m_threads.size(); }
	inline const WorkerThread *getThread(int index) const { return m_threads[index]; }
	inline WorkerThread *getThread(int index) { return m_threads[index]; }
	inline const Parameters &getParameters() const { return m_params; }

	inline void lock() {
		pthread_mutex_lock(&m_mutex);
	}

	inline void unlock() {
		pthread_mutex_unlock(&m_mutex);
	}

	inline void refreshViews() {
		pthread_mutex_lock(&m_refreshMutex);
		if (m_timer.getMilliseconds() > 100) {
			emit modelChangeEvent(Repaint);		
			m_timer.reset();
		}
		pthread_mutex_unlock(&m_refreshMutex);
	}

	inline bool getAngleVisibility(int angle) { return m_angleVisibility[angle]; }
	inline void setAngleVisibility(int angle, bool value) { 
		if (m_angleVisibility[angle] != value) {
			m_angleVisibility[angle] = value;
			emit modelChangeEvent(FocusAndRefresh);
		}
	}

	/* ==================================================================== */
	//    Utilities

	/* Convert a pixel from 3D pixel coodinates to the internal coordinate system */
	inline Point3 toPoint(const Point3s &pixel, int angle) const {
		const Float c = m_params.cosTheta[angle], s = m_params.sinTheta[angle];

		Point3 result(pixel.x + .5f - m_params.centerX,
					  pixel.y + .5f - m_params.yres/2.0,
					 -(pixel.z + .5f - m_params.centerDepth)/m_params.aspectZ);

		return Point3(c*result.x + s*result.z, result.y, -s*result.x + c*result.z);
	}

	/* Convert from the internal coordinate system to 3D pixel coordinates */
	inline Point3s toPixel(const Point3 &point, int angle) const {
		const Float c = m_params.cosTheta[angle], s = m_params.sinTheta[angle];

		Point3 rotPos(c*point.x - s*point.z, 
			point.y, s*point.x + c*point.z);

		return Point3s(
			(int) (rotPos.x + m_params.centerX),
			(int) (rotPos.y + m_params.yres/2.0),
			(int) (-rotPos.z * m_params.aspectZ + m_params.centerDepth)
		);
	}

	/* Compute the derivative of the projected X coordinate of the 
       given (focused) point as the camera moves back and forth */
	Vector2 dXYdZ(const Point3 &p, int angle) const {
		const Float c = m_params.cosTheta[angle], s = m_params.sinTheta[angle];
		const Float rotX = (c*p.x - s*p.z) * m_params.pixelWidth;
		const Float scaledY = p.y * m_params.pixelWidth, fp = m_params.fPlaneDist;
		return Vector2(-m_params.f*rotX, -m_params.f*scaledY) / (fp*fp);
	}

	/* As above, but returns a direction in world space */
	inline Vector3 dXYdZ_vec(const Point3 &p, int angle) const {
		const Float c = m_params.cosTheta[angle], s = m_params.sinTheta[angle];
		Vector2 dxy = dXYdZ(p, angle);
		Vector3 d = normalize(Vector3(dxy.x * m_params.pixelWidth, dxy.y*m_params.pixelWidth, 1));
		return Vector3(c*d.x + s*d.z, d.y, -s*d.x + c*d.z);
	}

	/* Get the projected 2D orientation of the given (position, direction) pair 
	   when it is focused from a particular angle */
	inline Float getOrientation(const Point3 &p_, const Vector3 &d, int angle) const {
		const Point3 p = p_ * m_params.pixelWidth;
		const Float c = m_params.cosTheta[angle], s = m_params.sinTheta[angle];
		const Float offset = -(s * p.x + c * p.z);

		Point3 p1(c*p.x - s*p.z, p.y, s*p.x + c*p.z);
		Point3 p2(c*(p.x+d.x) - s*(p.z+d.z), p.y + d.y, s*(p.x+d.x) + c*(p.z + d.z));

		Point2 p1Proj = Point2(p1.x, p1.y) / (p1.z + m_params.fPlaneDist + offset) * m_params.f;
		Point2 p2Proj = Point2(p2.x, p2.y) / (p2.z + m_params.fPlaneDist + offset) * m_params.f;

		return std::atan((p2Proj.y - p1Proj.y)/(p2Proj.x - p1Proj.x));
	}

	/* For a given position on a fiber observed with the supplied 2D orientation,
	   compute the normal of the 3D plane that both passes through the point and 
	   intersects the camera view plane with this orientation.
	*/
	Vector3 planeNormal(const Point3 &p, Float orientation, int angle) const {
		const Float c = m_params.cosTheta[angle], s = m_params.sinTheta[angle];

		/* First, convert to 3D pixel coordinates */
		const Point3s pixel = toPixel(p, angle);

		/* Compute 2 projected 2D positions on the fiber */
		Vector2 x1(pixel.x - m_params.centerX, pixel.y - m_params.yres/2);
		Vector2 x2 = x1 + Vector2(std::cos(orientation), std::sin(orientation));

		/* Move them into the local camera coordinate system using information when the point is in focus */
		Float dist = pixel.z * m_params.translateStep + m_params.translateStart + m_params.fPlaneDist;
		Point3 p1(x1.x/m_params.f*m_params.fPlaneDist, x1.y/m_params.f*m_params.fPlaneDist, m_params.fPlaneDist - dist);
		Point3 p2(x2.x/m_params.f*m_params.fPlaneDist, x2.y/m_params.f*m_params.fPlaneDist, m_params.fPlaneDist - dist);

		/* Compute the camera position as well as p1 and p2 in world space */
		Point3 p0 = Point3(-s*dist, 0, -c*dist);
		p1 = Point3(c*p1.x + s*p1.z, p1.y, -s*p1.x + c*p1.z);
		p2 = Point3(c*p2.x + s*p2.z, p2.y, -s*p2.x + c*p2.z);
    
		/* Normal of the plane passing through p0, p1 and p2 */
		return normalize(cross(p1-p0, p2-p0));
	}

	/**
	 * Constrain a vector to lie within a cone around another given vector
	 */
	Vector3 coneConstraint(const Vector3 &d, const Vector3 &axis, Float maxAngle) const {
		Float angle = std::acos(dot(d, axis));
		if (angle <= maxAngle)
			return d;
		Vector3 d2 = normalize(d - axis * dot(d, axis));
		return axis * std::cos(maxAngle) + d2 * std::sin(maxAngle);
	}

	/* Test whether a point in pixel coordinates is out of bounds */
	bool isValid(const Point3s &pixel) const {
		return (pixel.x >= 0 && pixel.x < m_params.xres &&
				pixel.y >= 0 && pixel.y < m_params.yres &&
				pixel.z >= 0 && pixel.z < m_params.numDepths);
	}
signals:
	void modelChangeEvent(DataModel::EChangeLevel level);
private:
	Parameters m_params;
	KDTree **m_sweeps;
	std::vector<HairFiber> m_fibers, m_comparisonFibers;
	std::vector<WorkerThread *> m_threads;
	std::vector<bool> m_angleVisibility;
	AABB m_aabb;
	bool m_updateViews;
	pthread_mutex_t m_mutex, m_refreshMutex;
	Timer m_timer;
};

#endif /* __DATAMODEL_H */
