#if !defined(__KDTREE_H)
#define __KDTREE_H

#include <hair/common.h>
#include <hair/obb.h>

/**
 * Hard-coded size limit (max 1,073,741,824) for recursion-free 
 * nearest neighbor searches.
 */
#define MAX_KDTREE_DEPTH 30

/** \brief Left-balanced KD-tree implementation for storing fiber observations
 * and executing efficient oriented box queries. All content is backed by
 * memory-mapped files, which makes it possible to work with extremely large 
 * datasets while alleviating the need of balancing the kd-tree on every 
 * program startup.
 * Based on Henrik Wann Jensen's photon map data structure. 
 *
 * @author Wenzel Jakob
 */
class KDTree {
public:
	struct Observation;

    /* ===================================================================== */
    /*                        Public access methods                          */
    /* ===================================================================== */

	/**
	 * This function does one of two things: when 'create' is set to true,
	 * the file given by the supplied filename is first created. Subsequent
	 * calls to storeObservation() will cause the observations to be written to disk. 
	 * After the last entry has been added, balance() builds the KD-tree and 
	 * makes it available for nearest-neighbor queries.
	 *
	 * When 'create' is set to false, the file is mmapped() into read-only memory
	 * and made available for nearest-neighbor queries.
	 *
	 * When the stored data has been scaled non-uniformly along the X, Y or Z
	 * axes, the 'scale' parameter can be used to transform it back.
	 */
	KDTree(const std::string &filename, const Vector3 &scale, bool create);

	/// Virtual destructor
	virtual ~KDTree();

	/// Return the axis-aligned box containing all observations
	inline AABB getAABB() const { 
		/* Convert from discrete pixel coordinates to intervals */
		return AABB(
			Point3(m_aabb.min.x, m_aabb.min.y, m_aabb.min.z),
			Point3(m_aabb.max.x+1, m_aabb.max.y+1, m_aabb.max.z+1)
		);
	}

	/// Return the total number of stored observations
	inline size_t getObservationCount() { return m_observationCount; }

	/// Look up an observation by index
	inline Observation &getObservation(size_t index) { return m_observations[index+1]; }
	
	/// Look up an observation by index (const version)
	inline const Observation &getObservation(size_t index) const { return m_observations[index+1]; }

	/**
	 * Store a number of observations in the (still unbalanced) kd-tree
	 */
	bool storeObservations(const std::vector<Observation> &observations);

	/**
	 * Search for observations contained in an oriented box
	 */
	void obbQuery(const OBB &obb, std::vector<Observation *> &results);

	/**
	 * Recursively build a left-balanced kd-tree. This has to be
	 * done once after all observations have been stored, but prior to
	 * executing any queries.
	 */
	void balance();

	/// Return a string representation
	std::string toString() const;
public:
    /* ===================================================================== */
    /*                        Shared data structures                         */
    /* ===================================================================== */

	/* Observation record (8 bytes). In addition to the 3D position and
	   the detected 2D orientation, a 'used' flag is also stored. This allows
	   features to be easily 'consumed' without removing them from the data structure.
	*/
	struct Observation {
		Point3s pos;			/* Observation position in 3D pixel coordinates */
		uint8_t splitAxis;		/* The name says it all */
		uint8_t flags;			/* Index into an array of orientations, as well as a 'used' flag */

		inline Point3 getPosition() const {
			return Point3(pos.x + .5f, pos.y + .5f, pos.z + .5f);
		}

		inline bool isUsed() const {
			return flags & 0x80;
		}

		inline void setUsed(bool active) {
			if (active)
				flags |= 0x80;
			else
				flags &= 0x7F;
		}

		inline Float getSplitPos() const {
			return pos[splitAxis] + .5f;
		}

		inline int getSplitAxis() const {
			return splitAxis;
		}

		inline uint8_t getOrientationIndex() const {
			return flags & 0x7F;
		}

		inline void setOrientationIndex(int idx) {
			flags = idx | (flags & 0x80);
		}

		inline Observation() : splitAxis(0), flags(0) { }

		inline Observation(const Point3s &p, uint8_t orientation) 
			: pos(p), splitAxis(0), flags(orientation) { }
	};
protected:
    /* ===================================================================== */
    /*                      Protected data structures                        */
    /* ===================================================================== */

	typedef Observation *observation_ptr;
	typedef const Observation *const_observation_ptr;
	typedef std::vector<observation_ptr>::iterator observation_iterator;

	/* Observation position comparison functor (< comparison). */
	struct functorLess : public std::unary_function<observation_ptr, bool> {
	public:
		inline functorLess(int splitAxis, short splitValue) 
			: axis(splitAxis), value(splitValue) {
		}

		inline bool operator()(observation_ptr observation) const {
			return observation->pos[axis] < value;
		}
	private:
		int axis;
		short value;
	};

	/* Observation position comparison functor (> comparison). */
	struct functorGreater : public 
		std::unary_function<observation_ptr, bool> {
	public:
		inline functorGreater(int splitAxis, short splitValue) 
			: axis(splitAxis), value(splitValue) {
		}

		inline bool operator()(observation_ptr observation) const {
			return observation->pos[axis] > value;
		}
	private:
		int axis;
		short value;
	};
protected:
    /* ===================================================================== */
    /*                         Protected subroutines                         */
    /* ===================================================================== */

	/* Short-valued 3D axis-aligned bounding box */
	struct AABB3s {
		Point3s min, max;
	
		inline AABB3s() {
			const short minValue = std::numeric_limits<Float>::min();
			const short maxValue = std::numeric_limits<Float>::max();
			min = Point3s(maxValue, maxValue, maxValue);
			max = Point3s(minValue, minValue, minValue);
		}
	
		inline void expandBy(const Point3s &vec) {
			min.x = std::min(min.x, vec.x);
			min.y = std::min(min.y, vec.y);
			min.z = std::min(min.z, vec.z);
			max.x = std::max(max.x, vec.x);
			max.y = std::max(max.y, vec.y);
			max.z = std::max(max.z, vec.z);
		}
	
		inline bool isValid() const {
			return max.x >= min.x && max.y >= min.y && max.z >= min.z;
		}
		
		inline std::string toString() const {
			std::ostringstream oss;
			oss << "AABB3s[";
			if (!isValid()) {
				oss << "invalid";
			} else {
				oss << "min=" << min.toString()
					<< ", max=" << max.toString();
			}
			oss	<< "]";
			return oss.str();
		}
	};

	/**
	 * Partition a list of observations so that there is an ordering between
	 * the observation at index 'pivot' and all other observations. See implementation
	 * for further documentation.
	 */
	void quickPartition(observation_iterator left, 
		observation_iterator right, observation_iterator pivot, int axis) const;

	/**
 	 * Given a number of entries, this method calculates the maximum amount of
	 * nodes on the left subtree of a left-balanced tree. See implementation
	 * for further documentation
	 */
	size_t leftSubtreeSize(size_t treeSize) const;

	/**
	 * Recursive balancing algorithm. See implementation for
	 * further documentation.
	 *
	 * @param sortStart
	 * 		The beginning of the list to be balanced
	 * @param sortEnd
	 * 		The end of the list to be balanced
	 * @param heapPointers
	 * 		An array, which will hold the observation tree as it is constructed
	 * 	@param heapIndex
	 * 		The heap array index of the next pivot
	 */
	void balanceRecursive(
		observation_iterator basePtr,
		observation_iterator sortStart,
		observation_iterator sortEnd,
		std::vector<size_t> &heapPermutation,
		AABB3s &aabb, size_t heapIndex) const;

	/// Heap access routines
	inline size_t leftChild(size_t index) const { return 2*index; }
	inline size_t rightChild(size_t index) const { return 2*index + 1; }
	inline bool isInnerNode(size_t index) const { return index <= m_lastInnerNode; }
	inline bool hasRightChild(size_t index) const { return index <= m_lastRChildNode; }
protected:
    /* ===================================================================== */
    /*                        Protected attributes                           */
    /* ===================================================================== */

	Observation *m_observations;
	AABB3s m_aabb;
	Vector3 m_scale;
	size_t m_observationCount;
	size_t m_lastInnerNode;
	size_t m_lastRChildNode;
	bool m_balanced;
	std::string m_filename;
	FILE *m_file;
	int m_fd;
	void *m_mmapPtr;
	int m_mmapSize;
};

#endif /* __KDTREE_H */
