#if !defined(__OBBTREE_H)
#define __OBBTREE_H

#include <hair/common.h>

/**
 * Oriented bounding box implementation as described in the paper
 * "OBB-Tree: A Hierarchical Structure for Rapid Interference 
 *  Detection (1996) S. Gottschalk, M.C. Lin, D. Manocha"
 *
 * @author Wenzel Jakob, 2008
 */
class OBB {
public:
	/// Construct an invalid OBB
	inline OBB() { }

	enum EClassification {
		ELeft,
		ERight,
		EBoth
	};

	/**
	 * Given a center, axes and extents along these axes, construct an OBB instance
	 */
	inline OBB(const Point3 &center, const Vector3 &extents, const Vector3 axes[3]) 
	 : m_center(center), m_extents(extents) {
		for (int i=0; i<3; ++i)
			m_axes[i] = axes[i];
	}

	/**
	 * Given a center, axes and extents along these axes, construct an OBB instance
	 */
	inline OBB(const Point3 &center, const Vector3 &extents, 
		const Vector3 axis0, const Vector3 axis1, const Vector3 axis2)
	 : m_center(center), m_extents(extents) {
		m_axes[0] = axis0;
		m_axes[1] = axis1;
		m_axes[2] = axis2;
	}

	/**
	 * Construct from an axis-aligned box
	 */
	inline OBB(const AABB &aabb) 
	 : m_center(aabb.getCenter()), m_extents(aabb.getExtents()/2) {
		m_axes[0] = Vector3(1, 0, 0);
		m_axes[1] = Vector3(0, 1, 0);
		m_axes[2] = Vector3(0, 0, 1);
	}

	/**
	 * Construct a OBB by running a principal component analysis on the input shape set.
	 * Only the range [rangeStart, rangeEnd) is analyzed.
	 */
	/*
	template <typename ShapeType> void PCA(const std::vector<ShapeType> &shapes, size_t rangeStart, size_t rangeEnd) {
		double cov[3][3];

		// Compute the covariance matrix 
		memset(cov, 0, sizeof(double)*3*3);
		for (size_t i=rangeStart; i<rangeEnd; ++i)
			m_center += shapes[i].getCenter();
		m_center /= rangeEnd-rangeStart;

		for (size_t k=rangeStart; k<rangeEnd; ++k) {
			Vector3 rel = shapes[k].getCenter() - m_center;
			for (int i=0; i<3; ++i)
				for (int j=0; j<3; ++j)
					cov[i][j] += rel[i]*rel[j];
		}

		// Retrieve its eigenvectors using GSL
		gsl_matrix_view m = gsl_matrix_view_array((double *) cov, 3, 3);
		gsl_vector *eval = gsl_vector_alloc(3);
		gsl_matrix *evec = gsl_matrix_alloc(3, 3);
		gsl_eigen_symmv_workspace *w = gsl_eigen_symmv_alloc(3);
		gsl_eigen_symmv(&m.matrix, eval, evec, w);
		gsl_eigen_symmv_free(w);
		gsl_eigen_symmv_sort(eval, evec, GSL_EIGEN_SORT_ABS_DESC);

		for (int i=0; i<3; ++i) {
			gsl_vector_view evec_i = gsl_matrix_column (evec, i);
			for (int j=0; j<3; ++j)
				m_axes[i][j] = gsl_vector_get(&evec_i.vector, j);
		}

		gsl_vector_free(eval);
		gsl_matrix_free(evec);

		/// Measure extents along the principal components
		for (size_t i=rangeStart; i<rangeEnd; ++i) {
			Vector3 v = shapes[i].getCenter() - m_center;
			Vector3 ext = shapes[i].getWorldExtents();
			for (int j=0; j<3; ++j)
				m_extents[j] = std::max(m_extents[j], absDot(v, m_axes[j])+ext[j]);
		}
	}
	*/

	/// Assignment operator
	inline void operator=(const OBB &obb) {
		m_center = obb.m_center;
		m_extents = obb.m_extents;
		for (int i=0; i<3; ++i)
			m_axes[i] = obb.m_axes[i];
	}

	/**
	 * Compute the OBB resulting from an affine transformation.
	 * The transformation is restricted to rotational+translational
	 * components.
	 */
	OBB transform(float A[4][4]) const;

	/// Return the center point
	inline const Point3 &getCenter() const { return m_center; }

	/**
	 * Compute the extents in world space. Measured from 
	 * the center (double to get the diameter)
	 */
	Vector3 getWorldExtents() const;

	/// Return the extents along one of the three OBB axes
	inline Float getExtents(int axis) const { return m_extents[axis]; }

	/// Return one of the OBB coordinate axes
	inline const Vector3 &getAxis(int axis) const { return m_axes[axis]; }
	
	/// Return one of the eight OBB corners
	Point3 getCorner(int corner) const;
	
	/// Test whether the OBB contains a point
	bool contains(const Point3 &p) const;
	
	/// Find the closest point on the OBB
	Point3 findClosest(const Point3 &p) const;

	/// Compute the distance of a point to this OBB
	inline Float getDistance(const Point3 &p) const {
		return (findClosest(p)-p).length();
	}

	/// Classify the OBB with respect to an axis-aligned splitting plane
	EClassification classify(int axis, Float value) const;

	/// Separating axis overlap test
	bool overlaps(const OBB &other) const;

	/// Draw a wireframe version using OpenGL
	void drawGL() const;

	/// Return a string representation
	std::string toString() const;
public:
	Point3 m_center;
	Vector3 m_extents, m_axes[3];
};

#endif /* __OBBTREE_H */
