#if !defined(__VECTOR_H)
#define __VECTOR_H

/* This file contains basic vector implementations of various sizes and types */

class Vector3 {
public:
	Float x, y, z;

	inline Vector3(Float _x = 0.0f, Float _y = 0.0f, Float _z = 0.0f)
		: x(_x), y(_y), z(_z) {
	}
	
	inline Vector3 operator+(const Vector3 &v) const {
		return Vector3(x + v.x, y + v.y, z + v.z);
	}

	inline Vector3 operator-(const Vector3 &v) const {
		return Vector3(x - v.x, y - v.y, z - v.z);
	}

	inline Vector3& operator+=(const Vector3 &v) {
		x += v.x; y += v.y; z += v.z;
		return *this;
	}

	inline Vector3& operator-=(const Vector3 &v) {
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}

	inline Vector3 operator*(Float f) const {
		return Vector3(x*f, y*f, z*f);
	}

	inline Vector3 &operator*=(Float f) {
		x *= f;
		y *= f;
		z *= f;
		return *this;
	}

	inline Vector3 operator-() const {
		return Vector3(-x, -y, -z);
	}

	inline Vector3 operator/(Float f) const {
		Float r = 1.0f / f;
		return Vector3(x * r, y * r, z * r);
	}

	inline Vector3 &operator/=(Float f) {
		Float r = 1.0f / f;
		x *= r;
		y *= r;
		z *= r;
		return *this;
	}

	inline Float operator[](int i) const {
		return (&x)[i];
	}

	inline Float &operator[](int i) {
		return (&x)[i];
	}

	inline Float lengthSquared() const {
		return x*x + y*y + z*z;
	}

	inline Float length() const {
		return std::sqrt(lengthSquared());
	}

	inline bool operator==(const Vector3 &v) const {
		return (v.x == x && v.y == y && v.z == z);
	}
	
	inline bool operator!=(const Vector3 &v) const {
		return !operator==(v);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "[" << x << ", " << y << ", " << z << "]";
		return oss.str();
	}
};

class Vector3i {
public:
	int x, y, z;

	inline Vector3i(int _x = 0, int _y = 0, int _z = 0)
		: x(_x), y(_y), z(_z) {
	}
	
	inline Vector3i operator+(const Vector3i &v) const {
		return Vector3i(x + v.x, y + v.y, z + v.z);
	}

	inline Vector3i operator-(const Vector3i &v) const {
		return Vector3i(x - v.x, y - v.y, z - v.z);
	}
	
	inline Vector3 operator-(const Vector3 &v) const {
		return Vector3(x - v.x, y - v.y, z - v.z);
	}

	inline Vector3i& operator+=(const Vector3i &v) {
		x += v.x; y += v.y; z += v.z;
		return *this;
	}

	inline Vector3i& operator-=(const Vector3i &v) {
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}

	inline Vector3i operator*(int f) const {
		return Vector3i(x*f, y*f, z*f);
	}

	inline Vector3i &operator*=(int f) {
		x *= f;
		y *= f;
		z *= f;
		return *this;
	}

	inline Vector3i operator-() const {
		return Vector3i(-x, -y, -z);
	}

	inline Vector3i operator/(int f) const {
		return Vector3i(x / f, y / f, z / f);
	}

	inline Vector3i &operator/=(int f) {
		x /= f;
		y /= f;
		z /= f;
		return *this;
	}

	inline int operator[](int i) const {
		return (&x)[i];
	}

	inline int &operator[](int i) {
		return (&x)[i];
	}

	inline int lengthSquared() const {
		return x*x + y*y + z*z;
	}

	inline int length() const {
		return std::sqrt(lengthSquared());
	}

	inline bool operator==(const Vector3i &v) const {
		return (v.x == x && v.y == y && v.z == z);
	}
	
	inline bool operator!=(const Vector3i &v) const {
		return !operator==(v);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "[" << x << ", " << y << ", " << z << "]";
		return oss.str();
	}
};

class Vector3s {
public:
	short x, y, z;

	inline Vector3s(short _x = 0, short _y = 0, short _z = 0)
		: x(_x), y(_y), z(_z) {
	}
	
	inline Vector3s operator+(const Vector3s &v) const {
		return Vector3s(x + v.x, y + v.y, z + v.z);
	}

	inline Vector3s operator-(const Vector3s &v) const {
		return Vector3s(x - v.x, y - v.y, z - v.z);
	}
	
	inline Vector3 operator-(const Vector3 &v) const {
		return Vector3(x - v.x, y - v.y, z - v.z);
	}

	inline Vector3s& operator+=(const Vector3s &v) {
		x += v.x; y += v.y; z += v.z;
		return *this;
	}

	inline Vector3s& operator-=(const Vector3s &v) {
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}

	inline Vector3s operator*(short f) const {
		return Vector3s(x*f, y*f, z*f);
	}

	inline Vector3s &operator*=(short f) {
		x *= f;
		y *= f;
		z *= f;
		return *this;
	}

	inline Vector3s operator-() const {
		return Vector3s(-x, -y, -z);
	}

	inline Vector3s operator/(short f) const {
		return Vector3s(x / f, y / f, z / f);
	}

	inline Vector3s &operator/=(short f) {
		x /= f;
		y /= f;
		z /= f;
		return *this;
	}

	inline short operator[](int i) const {
		return (&x)[i];
	}

	inline short &operator[](int i) {
		return (&x)[i];
	}

	inline short lengthSquared() const {
		return x*x + y*y + z*z;
	}

	inline short length() const {
		return std::sqrt(lengthSquared());
	}

	inline bool operator==(const Vector3s &v) const {
		return (v.x == x && v.y == y && v.z == z);
	}
	
	inline bool operator!=(const Vector3s &v) const {
		return !operator==(v);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "[" << x << ", " << y << ", " << z << "]";
		return oss.str();
	}
};

class Vector2s {
public:
	short x, y;

	inline Vector2s(short _x = 0, short _y = 0)
		: x(_x), y(_y) {
	}
	
	inline Vector2s operator+(const Vector2s &v) const {
		return Vector2s(x + v.x, y + v.y);
	}

	inline Vector2s operator-(const Vector2s &v) const {
		return Vector2s(x - v.x, y - v.y);
	}

	inline Vector2s& operator+=(const Vector2s &v) {
		x += v.x; y += v.y;
		return *this;
	}

	inline Vector2s& operator-=(const Vector2s &v) {
		x -= v.x; y -= v.y;
		return *this;
	}

	inline Vector2s operator*(short f) const {
		return Vector2s(x*f, y*f);
	}

	inline Vector2s &operator*=(short f) {
		x *= f;
		y *= f;
		return *this;
	}

	inline Vector2s operator-() const {
		return Vector2s(-x, -y);
	}

	inline Vector2s operator/(short f) const {
		short r = 1 / f;
		return Vector2s(x * r, y * r);
	}

	inline Vector2s &operator/=(short f) {
		short r = 1 / f;
		x *= r; y *= r;
		return *this;
	}

	inline short operator[](short i) const {
		return (&x)[i];
	}

	inline short &operator[](short i) {
		return (&x)[i];
	}

	inline short lengthSquared() const {
		return x*x + y*y;
	}

	inline short length() const {
		return (short) std::sqrt((Float) lengthSquared());
	}

	inline bool operator==(const Vector2s &v) const {
		return (v.x == x && v.y == y);
	}
	
	inline bool operator!=(const Vector2s &v) const {
		return !operator==(v);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "[" << x << ", " << y << "]";
		return oss.str();
	}
};

class Vector2i {
public:
	int x, y;

	inline Vector2i(int _x = 0, int _y = 0)
		: x(_x), y(_y) {
	}
	
	inline Vector2i operator+(const Vector2i &v) const {
		return Vector2i(x + v.x, y + v.y);
	}

	inline Vector2i operator-(const Vector2i &v) const {
		return Vector2i(x - v.x, y - v.y);
	}

	inline Vector2i& operator+=(const Vector2i &v) {
		x += v.x; y += v.y;
		return *this;
	}

	inline Vector2i& operator-=(const Vector2i &v) {
		x -= v.x; y -= v.y;
		return *this;
	}

	inline Vector2i operator*(int f) const {
		return Vector2i(x*f, y*f);
	}

	inline Vector2i &operator*=(int f) {
		x *= f;
		y *= f;
		return *this;
	}

	inline Vector2i operator-() const {
		return Vector2i(-x, -y);
	}

	inline Vector2i operator/(int f) const {
		int r = 1 / f;
		return Vector2i(x * r, y * r);
	}

	inline Vector2i &operator/=(int f) {
		int r = 1 / f;
		x *= r; y *= r;
		return *this;
	}

	inline int operator[](int i) const {
		return (&x)[i];
	}

	inline int &operator[](int i) {
		return (&x)[i];
	}

	inline int lengthSquared() const {
		return x*x + y*y;
	}

	inline int length() const {
		return (int) std::sqrt((Float) lengthSquared());
	}

	inline bool operator==(const Vector2i &v) const {
		return (v.x == x && v.y == y);
	}
	
	inline bool operator!=(const Vector2i &v) const {
		return !operator==(v);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "[" << x << ", " << y << "]";
		return oss.str();
	}
};

class Vector2 {
public:
	Float x, y;

	inline Vector2(Float _x = 0.0f, Float _y = 0.0f)
		: x(_x), y(_y) {
	}
	
	inline Vector2(const Vector2i &v) 
		: x((Float) v.x), y((Float) v.y) {
	}
	
	inline Vector2 operator+(const Vector2 &v) const {
		return Vector2(x + v.x, y + v.y);
	}

	inline Vector2 operator-(const Vector2 &v) const {
		return Vector2(x - v.x, y - v.y);
	}

	inline Vector2& operator+=(const Vector2 &v) {
		x += v.x; y += v.y;
		return *this;
	}

	inline Vector2& operator-=(const Vector2 &v) {
		x -= v.x; y -= v.y;
		return *this;
	}

	inline Vector2 operator*(Float f) const {
		return Vector2(x*f, y*f);
	}

	inline Vector2 &operator*=(Float f) {
		x *= f;
		y *= f;
		return *this;
	}

	inline Vector2 operator-() const {
		return Vector2(-x, -y);
	}

	inline Vector2 operator/(Float f) const {
		Float r = 1.0f / f;
		return Vector2(x * r, y * r);
	}

	inline Vector2 &operator/=(Float f) {
		Float r = 1.0f / f;
		x *= r; y *= r;
		return *this;
	}

	inline Float operator[](int i) const {
		return (&x)[i];
	}

	inline Float &operator[](int i) {
		return (&x)[i];
	}

	inline Float lengthSquared() const {
		return x*x + y*y;
	}

	inline Float length() const {
		return std::sqrt(lengthSquared());
	}

	inline bool operator==(const Vector2 &v) const {
		return (v.x == x && v.y == y);
	}
	
	inline bool operator!=(const Vector2 &v) const {
		return !operator==(v);
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "[" << x << ", " << y << "]";
		return oss.str();
	}
};

typedef Vector3 Point3;
typedef Vector3s Point3s;
typedef Vector2s Point2s;
typedef Vector2 Point2;

inline Float dot(const Vector3 &v1, const Vector3 &v2) {
	return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
}

inline Float absDot(const Vector3 &v1, const Vector3 &v2) {
	return std::abs(v1.x*v2.x + v1.y*v2.y + v1.z*v2.z);
}

inline Float dot(const Vector2 &v1, const Vector2 &v2) {
	return v1.x*v2.x + v1.y*v2.y;
}

inline Float absDot(const Vector2 &v1, const Vector2 &v2) {
	return std::abs(v1.x*v2.x + v1.y*v2.y);
}

inline Vector3 normalize(const Vector3 &v) {
	return v / std::sqrt(dot(v,v));
}

inline Vector2 normalize(const Vector2 &v) {
	return v / std::sqrt(dot(v,v));
}

inline Vector3 cross(const Vector3 &v1, const Vector3 &v2) {
	/* Left-handed vector cross product */
	return Vector3(
		(v1.y * v2.z) - (v1.z * v2.y), 
		(v1.z * v2.x) - (v1.x * v2.z),
		(v1.x * v2.y) - (v1.y * v2.x)
	);
}

inline void coordinateSystem(const Vector3 &a, Vector3 &b, Vector3 &c) {
	if (std::abs(a.x) > std::abs(a.y)) {
		Float invLen = 1.0f / std::sqrt(a.x * a.x + a.z * a.z);
		b = Vector3(-a.z * invLen, 0.0f, a.x * invLen);
	} else {
		Float invLen = 1.0f / std::sqrt(a.y * a.y + a.z * a.z);
		b = Vector3(0.0f, -a.z * invLen, a.y * invLen);
	}
	c = cross(a, b);
}

struct Ray2 {
	Point2 p;
	Vector2 d;

	inline Ray2() {
	}

	inline Ray2(const Point2 &p, const Vector2 &d) 
	 : p(p), d(d) {
	}
};

struct Ray3 {
	Point3 o;
	Vector3 d;
	Float mint, maxt;

	inline Ray3() : mint(Epsilon), maxt(std::numeric_limits<Float>::infinity()) {
	}

	inline Ray3(const Point3 &o, const Vector3 &d) 
	 : o(o), d(d), mint(Epsilon), maxt(std::numeric_limits<Float>::infinity()) {
	}

	inline Ray3(const Ray3 &ray) : o(ray.o), d(ray.d), mint(ray.mint), maxt(ray.maxt) {
	}

	inline std::string toString() const {
		std::ostringstream oss;
		oss << "Ray3[o=" << o.toString() << ", d=" << d.toString() << "]";
		return oss.str();
	}

	inline Point3 operator()(Float t) const {
		return o + d*t;
	}
};

#endif /* __VECTOR_H */
