#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/bindings/traits/ublas_matrix.hpp>
#include <boost/numeric/bindings/traits/ublas_vector.hpp>
#include <boost/numeric/bindings/lapack/gesvd.hpp>
#include <boost/numeric/bindings/lapack/gels.hpp>
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <hair/datamodel.h>
#include <hair/wthread.h>
#include <hair/glwidget.h>
#include <fstream>

namespace ublas  = boost::numeric::ublas;
namespace lapack = boost::numeric::bindings::lapack;

DataModel::DataModel() : m_sweeps(NULL) {
	m_updateViews = true;
	pthread_mutex_init(&m_mutex, NULL);
	pthread_mutex_init(&m_refreshMutex, NULL);
}

DataModel::~DataModel() {
	pthread_mutex_destroy(&m_mutex);
	pthread_mutex_destroy(&m_refreshMutex);
	if (m_sweeps) {
		for (int i=0; i<m_params.numAngles; ++i)
			delete m_sweeps[i];
		delete[] m_sweeps;
	}
}

void DataModel::load(const Parameters &params) {
	if (m_sweeps) {
		for (int i=0; i<m_params.numAngles; ++i)
			delete m_sweeps[i];
		delete[] m_sweeps;
	}

	m_params = params;
	m_params.precompute();
	m_sweeps = new KDTree*[m_params.numAngles];
	m_aabb.reset();
	m_angleVisibility.clear();

	for (size_t i=0; i<m_threads.size(); ++i)
		delete m_threads[i];
	m_threads.clear();

	for (int angle=0; angle<params.numAngles; ++angle) {
		std::string outputFilename = formatString(params.filterOutput.c_str(), angle);
		cout << "  - Mapping \"" << outputFilename << "\" into memory .." << endl;
		KDTree *tree = new KDTree(outputFilename, Vector3(1, 1, 1/params.aspectZ), false);
		for (size_t i=0; i<tree->getObservationCount(); ++i) {
			KDTree::Observation &observation = tree->getObservation(i);
			if (observation.isUsed())
				observation.setUsed(false);
			m_aabb.expandBy(toPoint(observation.pos, angle));
		}
		m_sweeps[angle] = tree;
		m_angleVisibility.push_back(false);
	}
	
	int sideCount = std::sqrt(params.threadCount);
	if (sideCount * sideCount != params.threadCount)
		throw std::runtime_error("Thread count must be a square of an integer!");

	Vector3 chunkExtents = m_aabb.getExtents() / (Float) sideCount;
	for (int i=0; i<params.threadCount; ++i) {
		int x = i%sideCount;
		int z = i/sideCount;

		AABB aabb;
		aabb.min.x = m_aabb.min.x + chunkExtents.x * x;
		aabb.min.z = m_aabb.min.z + chunkExtents.z * z;
		aabb.min.y = m_aabb.min.y;

		aabb.max.x = aabb.min.x + chunkExtents.x;
		aabb.max.z = aabb.min.z + chunkExtents.z;
		aabb.max.y = m_aabb.max.y;

		m_threads.push_back(new WorkerThread(aabb, this));
	}

	emit modelChangeEvent(NewFile);
}

const ActiveSet &DataModel::getActiveSet(int index) const {
	return static_cast<WorkerThread *>(m_threads[index / 2])->getActiveSet(index % 2);
}

void DataModel::obbQuery(const OBB &obb, int angle, std::vector<KDTree::Observation *> &result) const {
	const Float c = m_params.cosTheta[angle], 
				s = m_params.sinTheta[angle];
	Float m[4][4] = {
		{c,	0, -s,   m_params.centerX},
		{0,	1,  0,   m_params.yres/2.0f},
		{-s, 0, -c,  m_params.centerDepth/m_params.aspectZ},
		{0,	0,  0,   1}
	};

	OBB transformed = obb.transform(m);

	m_sweeps[angle]->obbQuery(transformed, result);
}

struct OrientationTest {
	OrientationTest(Float orientation, Float delta, 
		const Parameters &parameters) : orientation(orientation), 
		delta(delta), parameters(parameters) {
	}

	inline bool operator() (KDTree::Observation *obs) {
		Float value = parameters.ridgeOrientations[obs->getOrientationIndex()];
		Float difference = (value-orientation)*180/M_PI;
		while (difference < -90)
			difference += 180;
		while (difference > 90)
			difference -= 180;

		return std::abs(difference) > delta;
	}

	Float orientation, delta;
	const Parameters &parameters;
};

void DataModel::obbQuery(const OBB &obb, Float orientation, Float delta, 
		int angle, std::vector<KDTree::Observation *> &result) const {
	obbQuery(obb, angle, result);
	result.erase(std::remove_if(result.begin(), result.end(), 
		OrientationTest(orientation, delta, m_params)), result.end());
}

/* SVD-based direction fitting */
bool DataModel::solveDirection(ActiveSet &activeSet, Vector3 &newDirection) const {
	using namespace ublas;
	bool repeat = true;
	Vector3 oldDirection = activeSet.d;

	do {
		if (activeSet.nActive < m_params.minObservations) {
			return false;
		}

		matrix<Float, column_major> A(activeSet.nActive,3), A2(activeSet.nActive, 3);

		int pos = 0;
		for (int angle=0; angle<MAX_ACTIVESET_SIZE; ++angle) {
			if (!activeSet.active[angle])
				continue;
			Vector3 n = planeNormal(activeSet.p, activeSet.orientation[angle], angle);
			A(pos, 0) = n.x;
			A(pos, 1) = n.y;
			A(pos, 2) = n.z;
			++pos;
		}

		matrix<Float, column_major> U(activeSet.nActive, activeSet.nActive);
		matrix<Float, column_major> V(3,3);
		ublas::vector<Float> S(std::min(3, activeSet.nActive));

		A2 = A; /* GESVD overwrites A */
		lapack::gesvd('O', 'N', 'A', A, S, U, V); // Compute all of V^T
		ublas::vector<Float> dir = trans(row(V, 2));
		ublas::vector<Float> errVec = prod(A2, dir);
		newDirection = Vector3(dir(0), dir(1), dir(2));

		/* Discard outliers */
		pos = 0;
		repeat = false;
		for (int i=0; i<m_params.numAngles; ++i) {
			if (!activeSet.active[i])
				continue;
			if (std::abs(errVec(pos++)) > MAX_FIT_ERROR) {
				activeSet.active[i] = false;
				activeSet.nActive--;
				repeat = true;
			}
		}
	} while (repeat);

	if (dot(newDirection, oldDirection) < 0)
		newDirection = -newDirection;

	return true;
}

/* Re-centering correction */
bool DataModel::solveCorrection(ActiveSet &activeSet, const Point3 &proposed, Vector3 &correction) const {
	using namespace ublas;
	std::vector<KDTree::Observation *> result;

	Vector3 X, Y;
	coordinateSystem(activeSet.d, X, Y);
	std::vector<Vector2> normals;
	std::vector<Float> rhs;

	int nObservations = 0;
	for (int angle=0; angle<m_params.numAngles; ++angle) {
		Vector3 dxdz = dXYdZ_vec(proposed, angle);
	
		/* Orthogonalize wrt. current direction */
		dxdz = normalize(dxdz - activeSet.d*dot(dxdz, activeSet.d));

		result.clear();
		Float orientation = getOrientation(proposed, activeSet.d, angle);
		obbQuery(OBB(proposed, m_params.sheetSize, dxdz, activeSet.d, cross(dxdz, activeSet.d)), 
			orientation, m_params.angleThreshold, angle, result);

		/* Compute 2D normal */
		Vector2 n = Vector2(-dot(Y, dxdz), dot(X, dxdz));

		/* Project onto X, Y */
		for (size_t j=0; j<result.size(); ++j) {
			Point3 worldPos = toPoint(result[j]->pos, angle);
			Vector2 rel = Vector2(dot(X, worldPos-proposed), 
				dot(Y, worldPos-proposed));

			rhs.push_back(dot(rel, n));
			normals.push_back(n);
		}

		if (result.size() > 0) 
			++nObservations;
	}

	if (nObservations < m_params.minObservations) 
		return false;

	/* Pack the data into a form that GELS handles */
	matrix<Float, column_major> A(normals.size(), 2);
	ublas::vector<Float> b(rhs.size());
	for (size_t i=0; i<rhs.size(); ++i) {
		A(i,0) = normals[i].x;
		A(i,1) = normals[i].y;
		b(i) = rhs[i];
	}

	lapack::gels('N', A, b, lapack::optimal_workspace());

	correction = X*b(0) + Y*b(1);

	return true;
}

/* Acquire updated 2D orientations after a step has been taken */
void DataModel::findOrientations(ActiveSet &activeSet) const {
	std::vector<KDTree::Observation *> result;

	/* Use the unconstrained directions/position data for the lookup */
	Vector3 d = activeSet.dOriginal;
	Point3 p = activeSet.pOriginal;

	Vector3 X, Y;
	coordinateSystem(d, X, Y);

	activeSet.lock();
	activeSet.nActive = 0;

	/* Search for close-by features to determine on which angles this hair is visible */
	for (int angle=0; angle<m_params.numAngles; ++angle) {
		Vector3 dxdz = dXYdZ_vec(p, angle);

		/* Orthogonalize wrt. current direction */
		dxdz = normalize(dxdz - d*dot(dxdz, activeSet.d));

		result.clear();
		Float orientation = getOrientation(p, d, angle);
		obbQuery(OBB(p, m_params.sheetSize, dxdz, d, cross(dxdz, d)), 
			orientation, m_params.angleThreshold, angle, result);

		/* Feature direction in the 2D subspace orthogonal to the growing direction */
		Vector2 d = Vector2(
			dot(X, dxdz),
			dot(Y, dxdz)
		);

		Vector2 direction(0,0);

		int numMatched = 0;
		for (size_t j=0; j<result.size(); ++j) {
			Point3 worldPos = toPoint(result[j]->pos, angle);
			Vector2 rel = Vector2(dot(X, worldPos-p), dot(Y, worldPos-p));

			// how far is the feature away from being in focus?
			Float absDist = std::abs(dot(rel, d));

			/* Consider only reasonably focused observations */
			if (absDist <= 7) {
				Float orientation = m_params.ridgeOrientations[result[j]->getOrientationIndex()];
				Vector2 d(std::cos(orientation), std::sin(orientation));
				if (dot(d, direction) < 0)
					d = -d;
				direction += d;
				numMatched++;
			}

			activeSet.observations[angle].insert(result[j]);
		}

		if (numMatched > 0) {
			activeSet.orientation[angle] = std::atan(direction.y/direction.x);
			activeSet.active[angle] = true;
			activeSet.nActive++;
		} else {
			activeSet.active[angle] = false;
		}
	}

	activeSet.unlock();
}

bool DataModel::grow(ActiveSet &activeSet) {
	if (activeSet.blindSteps >= MAX_BLIND_STEPS)
		return false;
	for (int i=0; i<m_params.numAngles; ++i)
		activeSet.fiber.seen[i] |= activeSet.active[i];

	bool blind = false;
	Vector3 prevDir = activeSet.d;

	findOrientations(activeSet);

	Vector3 prelimDirection;
	if (solveDirection(activeSet, prelimDirection)) {
		/* Enlarge the window of acceptable angles depending on the number of preceding blind steps */
		Float cutoff = m_params.directionCutoff * (activeSet.blindSteps + 1);
		activeSet.d = coneConstraint(prelimDirection, activeSet.d, cutoff);
		activeSet.dOriginal = prelimDirection;
	} else {
		blind = true;
	}

	activeSet.pOriginal = activeSet.p + activeSet.d * GROW_STEPSIZE;
	Vector3 prelimCorrection;
	if (solveCorrection(activeSet, activeSet.pOriginal, prelimCorrection)) {
		/* Relax constraints when blindly following a direction */
		Float maxMagnitude = m_params.correctionCutoff * (activeSet.blindSteps + 1);
		if (prelimCorrection.length() > maxMagnitude) {
			prelimCorrection = normalize(prelimCorrection) * maxMagnitude;
		}
		activeSet.p = activeSet.pOriginal + prelimCorrection;
	} else {
		blind = true;
		activeSet.p = activeSet.pOriginal;
	}

	activeSet.lock();
	activeSet.fiber.vertices.push_back(activeSet.p);
	activeSet.nSteps++;

	if (blind) {
		activeSet.blindSteps++;
	} else if (activeSet.blindSteps > 0) {
		activeSet.blindSteps = 0;
	}

	activeSet.unlock();
	refreshViews();

	return true;
}


void DataModel::go() {
	for (size_t i=0; i<m_threads.size(); ++i) {
		WorkerThread *thread = static_cast<WorkerThread *>(m_threads[i]);
		if (thread->isActive())
			thread->stop();
		else
			thread->start();
	}
}

void DataModel::addFiber(const HairFiber &fiber) {
	lock();
	m_fibers.push_back(fiber);
	unlock();
}

void DataModel::clearFibers() {
	lock();
	m_fibers.clear();
	m_comparisonFibers.clear();
	unlock();
}

void DataModel::loadMDL(const std::string &filename, bool comparison) {
	std::ifstream is(filename.c_str());
	std::string line;
	HairFiber fiber;
	bool pstnTag = false;
	bool vis[MAX_ACTIVESET_SIZE];
	for (int i=0; i<MAX_ACTIVESET_SIZE; ++i)
		vis[i] = 1; /* Compatibility to older datasets */

	while (!is.eof()) {
		std::getline(is, line);
		if (line == "  vrtxPstn") {
			fiber = HairFiber();
			pstnTag = true;
		} else if (line == "  end") {
			for (int i=0; i<MAX_ACTIVESET_SIZE; ++i)
				fiber.seen[i] = vis[i];
			if (fiber.vertices.size() > 0) {
				if (comparison)
					m_comparisonFibers.push_back(fiber);
				else
					m_fibers.push_back(fiber);
			}
			pstnTag = false;
  		} else if (strncmp(line.c_str(), "  # VIS: ", 9)==0) {
			line = line.substr(9);
			line = line.substr(0, line.length()-1);
			boost::char_separator<char> sep(" ");
			boost::tokenizer<boost::char_separator<char> > tokens(line, sep);

			int pos = 0;
			for (boost::tokenizer<boost::char_separator<char> >::const_iterator
					it = tokens.begin(); it != tokens.end(); ++it) {
				const std::string token = *it;
				vis[pos++] = boost::lexical_cast<int>(token);
			}
		} else if (pstnTag) {
			float f1, f2, f3;
			sscanf(line.c_str(), "	%f %f %f", &f1, &f2, &f3);
			fiber.vertices.push_back(Point3(f1, f2, f3));
		}
	}

	is.close();
	emit modelChangeEvent(Refresh);
}

void DataModel::saveOBJ(const std::string &filename) {
	std::ofstream os(filename.c_str());
	for (size_t i=0; i<m_fibers.size(); ++i) {
		const HairFiber &fiber = m_fibers[i];
	
		for (size_t j=0; j<fiber.vertices.size(); ++j) {
			const Point3 &p = fiber.vertices[j];
			os << "v " << p.x << " " << p.y << " " << p.z << endl;
		}
	}

	int idx = 1;
	for (size_t i=0; i<m_fibers.size(); ++i) {
		const HairFiber &fiber = m_fibers[i];

		for (size_t j=0; j<fiber.vertices.size()-1; ++j) {
			os << "l " << idx << " " << idx+1 << endl;
			++idx;
		}
		++idx;
	}
	os.close();
}

void DataModel::saveMDL(const std::string &filename, const std::vector<HairFiber> &fibers) const {
	std::ofstream os(filename.c_str());
	for (size_t i=0; i<fibers.size(); ++i) {
		const HairFiber &fiber = fibers[i];
	
		os << "hr \"hair" << i << "\"" << endl;
		os << "  # VIS: ";
		for (int i=0; i<m_params.numAngles; ++i)
			os << (int) fiber.seen[i] << " ";
		os << endl;
		os << "  mtrlNm \"hairMtrl\" end" << endl;
		os << "  " << 0.041 / m_params.pixelWidth << endl;
		os << "  vrtxPstn" << endl;

		for (size_t j=0; j<fiber.vertices.size(); ++j) {
			const Point3 &p = fiber.vertices[j];
			os << "	" << p.x << " " << p.y << " " << p.z << endl;
		}
		os << "  end" << endl;
		os << "end" << endl;
	}

	os.close();
}
