#include <hair/filter.h>
#include <hair/timer.h>
#include <fstream>
#include <omp.h>
#include <cstdlib>

static CvMat **filters;

/**
 * Construct a gabor filter where
 * 	
 * @param theta  rotates the filter
 * @param sigmaX controls the gaussian envelope (horizontally)
 * @param lambda determines the frequency
 * @param gamma  specifies the aspect ratio
 * @param psi    shifts the phase
 */
CvMat *gaborFilter(float theta, float sigmaX = 1, float lambda = 3, float gamma = 1, float psi = 0) {
	float sigmaY = sigmaX / gamma;
	int sizeX = (int) (7 * sigmaX);
	int sizeY = (int) (7 * sigmaY);
	if (sizeX % 2 == 0) ++sizeX;
	if (sizeY % 2 == 0) ++sizeY;

	CvMat *filter = cvCreateMat(sizeX, sizeY, CV_32FC1);

	sigmaX *= sigmaX;
	sigmaY *= sigmaY;

	float floatSum = 0;

	for (int y = -sizeY/2; y<=sizeY/2; ++y) {
		for (int x = -sizeX/2; x<=sizeX/2; ++x) {
			float xRot =  x*std::cos(theta)+y*std::sin(theta);
			float yRot = -x*std::sin(theta)+y*std::cos(theta);

			float envelope = std::exp(-0.5f * ((xRot*xRot)/sigmaX + (yRot*yRot)/sigmaY));
			float harmonic = std::cos(2*(float)M_PI/lambda*xRot+psi);
			floatSum += harmonic*envelope;
		}
	}
	floatSum /= sizeX*sizeY;

	for (int y = -sizeY/2; y<=sizeY/2; ++y) {
		for (int x = -sizeX/2; x<=sizeX/2; ++x) {
			float xRot =  x*std::cos(theta)+y*std::sin(theta);
			float yRot = -x*std::sin(theta)+y*std::cos(theta);

			float envelope = std::exp(-0.5f * ((xRot*xRot)/sigmaX + (yRot*yRot)/sigmaY));
			float harmonic = std::cos(2*(float)M_PI/lambda*xRot+psi);
			float value = envelope * harmonic - floatSum;
			cvSetReal2D(filter, y+sizeY/2, x+sizeX/2, value);
		}
	}

	return filter;
}

/**
 * Recursive hysteresis thresholding implementation
 * 
 * @param src 
 *     Gabor-filtered input input image
 * @param orientations
 *     Associated orientation field
 * @param pixel
 *     Pixel used to start the hysteresis thresholding operation
 * @param observations
 *     Output data structure for the observations
 */
void hysteresisThreshold(IplImage *src, uint8_t *orientations, 
		const Point3s &pixel, std::vector<KDTree::Observation> &observations,
		const Parameters &params) {
	short *data = (short *)(src->imageData + src->widthStep*pixel.y);
	uint8_t orientation = orientations[params.xres * pixel.y + pixel.x];

	short strength = data[pixel.x];
	if (!strength)
		return;

	data[pixel.x] = 0;

	for (int yo=-1; yo<=1; ++yo) {
		for (int xo=-1; xo<=1; ++xo) {
			if (xo == 0 && yo == 0)
				continue;

			if (pixel.x+xo < 1 || pixel.x+xo >= src->width-1 || 
				pixel.y+yo < 1 || pixel.y+yo >= src->height-1)
				continue;
	
			short other = ((short *)(src->imageData + src->widthStep*(pixel.y+yo)))[pixel.x + xo];
			if (other)
				hysteresisThreshold(src, orientations,
					Point3s(pixel.x+xo, pixel.y+yo, pixel.z), observations, params);
		}
	}

	observations.push_back(KDTree::Observation(pixel, orientation));
}

/* Filter an image using a Gabor filter bank & apply hysteresis thresholding.
   Returns a vector of observations in the form of oriented pixels */
void findHairs(IplImage *input, IplImage *debugImage, int depth, std::vector<KDTree::Observation> &result,
		const Parameters &parameters) {
	CvMat **filtered = new CvMat*[parameters.gaborCount];
	IplImage *src = NULL;

	/* Scale-space framework: if requested, start by convolving with a Gaussian filter kernel */
	if (parameters.gaussianExtents > 1) {
		cout << "    + Smoothing .." << endl;
		src = cvCreateImage(cvSize(input->width, input->height), IPL_DEPTH_16S, 1);
		cvSmooth(input, src, CV_GAUSSIAN, parameters.gaussianExtents);
	} else {
		src = input;
	}

	cout << "    + Detecting ridge orientations .. (" << omp_get_num_procs() << " threads)"<< endl;

	/* Apply the Gabor filter kernels in parallel */
	#pragma omp parallel
	{
		#pragma omp for schedule(dynamic, 1) 
		for (int i=0; i<parameters.gaborCount; ++i) {
			filtered[i] = cvCreateMat(src->height, src->width, CV_16SC1);
			cvFilter2D(src, filtered[i], filters[i]);
		}
	}

	IplImage *dest   = cvCreateImage(cvSize(src->width, src->height), IPL_DEPTH_16S, 1);
	cvZero(dest);
	uint8_t *destOrientations = new uint8_t[src->width*src->height];
	memset(destOrientations, 0, src->width*src->height*sizeof(uint8_t));
	cout << "    + Non-maximum suppression .." << endl;

	if (debugImage) {
		int minValue = INT_MAX, maxValue = INT_MIN;
		for (int y=0; y<src->height; ++y) {
			const short *L = (short *) (src->imageData + src->widthStep*y);
			uint8_t *D = (uint8_t *) (debugImage->imageData + debugImage->widthStep*y);
			for (int x=0; x<src->width; ++x) {
				minValue = std::min((int) L[x], minValue);
				maxValue = std::max((int) L[x], maxValue);
				for (int i=0; i<3; ++i) {
					D[3*x+i] = std::min((L[x]-130)/3,255);
				}
			}
		}
	}

	/* Perform non-maximum suppression in parallel */
	#pragma omp parallel
	{
		#pragma omp for schedule(dynamic, 50) 
		for (int y=1; y<src->height-1; ++y) {
			const short *L   = (short *) (src->imageData + src->widthStep*y);
			short *d  = (short *) (dest->imageData  + dest->widthStep*y);

			for (int x=1; x<src->width-1; ++x) {
				int maxResponse = INT_MIN, maxResponseIndex = -1;
				for (int i=0; i<parameters.gaborCount; ++i) {
					int response = (int) cvGetReal2D(filtered[i], y, x);
					if (response > maxResponse) {
						maxResponse = response;
						maxResponseIndex = i;
					}
				}

				if (maxResponseIndex == -1)
					throw std::logic_error("Internal error while filtering!");

				const float maxResponseAngle = parameters.filterOrientations[maxResponseIndex],
					xAxisMax = std::cos(maxResponseAngle),
					yAxisMax = std::sin(maxResponseAngle);

				const int M  = L[x];
				const int M1Max = cvGetInterpolated16(src, y+yAxisMax, x+xAxisMax);
				const int M2Max = cvGetInterpolated16(src, y-yAxisMax, x-xAxisMax);
				const bool hasMax = M > M1Max && M > M2Max;
				Float surroundMax = std::max(M1Max, M2Max);

				Float relVal = (M - surroundMax) / 
					(surroundMax + parameters.filterOffset) * 1000;

				if (hasMax && relVal > parameters.hysteresisLow) {
					d[x] = (int) relVal;
					destOrientations[src->width*y + x] = (uint8_t) maxResponseIndex;
				} else {
					d[x] = 0;
				}
			}
		}
	}

	cout << "    + Hysteresis thresholding .." << endl;
	for (int y=1; y<src->height-1; ++y) {
		for (int x=1; x<src->width-1; ++x) {
			short magnitude = ((short *)(dest->imageData + dest->widthStep*y))[x];
			if (magnitude < parameters.hysteresisHigh)
				continue;

			hysteresisThreshold(dest, destOrientations, Point3s(x,y, depth), result, parameters);
		}
	}
			
	if (debugImage) {
		for (size_t i=0; i<result.size(); ++i) 
			cvSet2D(debugImage, result[i].pos.y, result[i].pos.x, cvScalar(50, 50, 255));
	}

	cvReleaseImage(&dest);
	delete[] destOrientations;

	for (int i=0; i<parameters.gaborCount; ++i) 
		cvReleaseMat(&filtered[i]);
			
	if (parameters.gaussianExtents > 1)
		cvReleaseImage(&src);

	delete[] filtered;
}

/* Run the whole filter stage as configured by the 'parameters' data structure */
void runFilterStage(const Parameters &parameters) {
	int numRemaining = parameters.numAngles*parameters.numDepths;
	int numDone = 0;
	Timer timer;

	cout << " - Constructing Gabor filter bank .." << endl;
	filters = new CvMat*[parameters.gaborCount];
	for (int i=0; i<parameters.gaborCount; ++i) 
		filters[i] = gaborFilter(parameters.filterOrientations[i]);

	for (int angle=0; angle<parameters.numAngles; ++angle) {
		std::string outputFilename = formatString(parameters.filterOutput.c_str(), angle);
		KDTree kdTree(outputFilename, Vector3(1, 1, 1/parameters.aspectZ), true);
		for (int depth=0; depth<parameters.numDepths; ++depth) {
			cout << " - Processing angle " << angle << " / depth " << depth << " (ETA:" 
				<< timeToString(((timer.getMilliseconds() / (float) numDone) * numRemaining) / 1000.0f)
				<< ")" << endl;
			int index = angle*parameters.numDepths+depth;
			std::string inputFilename  = formatString(parameters.input.c_str(), index);
			std::string debugFilename = formatString("debug-%i.png", index);
			IplImage *src = cvLoadPNG16(inputFilename), *debug = NULL;

#ifdef DEBUG
			debug = cvCreateImage(cvSize(src->width, src->height), IPL_DEPTH_8U, 3);
#endif

			std::vector<KDTree::Observation> observations;
			findHairs(src, debug, depth, observations, parameters);
			kdTree.storeObservations(observations);
			cout << "    + Recorded " << observations.size() << " hair observations .." << endl;
			cout << endl;
			++numDone; --numRemaining;
			cvReleaseImage(&src);
#ifdef DEBUG
			cvSavePNG8(debugFilename, debug);
#endif
		}
		kdTree.balance();
	}

	cout << " - Releasing Gabor filter bank .." << endl;
	for (int i=0; i<parameters.gaborCount; ++i)
		cvReleaseMat(&filters[i]);
	delete[] filters;
}

