#include <hair/glviewer.h>
#include <hair/wthread.h>
#include <hair/filter.h>
#include <hair/timer.h>
#include <fstream>

GLViewer::GLViewer(const Parameters &params) {
	setWindowTitle(tr("OpenGL hair viewer"));
	
	QAction *exitAct = new QAction(tr("E&xit"), this);
	exitAct->setShortcut(tr("Ctrl+Q"));
	connect(exitAct, SIGNAL(triggered()), this, SLOT(close()));
	
	QAction *clearFibersAct = new QAction(tr("Clear fibers"), this);
	connect(clearFibersAct, SIGNAL(triggered()), this, SLOT(clearFibers()));
	QAction *loadMDLAct = new QAction(tr("Load fibers from &MDL.."), this);
	connect(loadMDLAct, SIGNAL(triggered()), this, SLOT(loadMDL()));
	QAction *loadMDLComparisonAct = new QAction(tr("Load comparison from &MDL.."), this);
	connect(loadMDLComparisonAct, SIGNAL(triggered()), this, SLOT(loadMDLComparison()));
	QAction *saveMDLAct = new QAction(tr("Save fibers to &MDL.."), this);
	connect(saveMDLAct, SIGNAL(triggered()), this, SLOT(saveMDL()));
	QAction *saveOBJAct = new QAction(tr("Save fibers to &OBJ.."), this);
	connect(saveOBJAct, SIGNAL(triggered()), this, SLOT(saveOBJ()));

	QMenu *fileMenu = new QMenu(tr("&File"), this);
	fileMenu->addAction(clearFibersAct);
	fileMenu->addSeparator();
	fileMenu->addAction(loadMDLAct);
	fileMenu->addAction(loadMDLComparisonAct);
	fileMenu->addAction(saveMDLAct);
	fileMenu->addAction(saveOBJAct);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAct);

	QAction *videoResAct = new QAction(tr("Set video resolution"), this);
	connect(videoResAct, SIGNAL(triggered()), this, SLOT(videoResolution()));

	m_updateViewsAct = new QAction(tr("Auto-update views"), this);
	m_updateViewsAct->setCheckable(true);
	m_updateViewsAct->setChecked(true);
	connect(m_updateViewsAct, SIGNAL(triggered()), this, SLOT(updateViews()));
	
	QMenu *viewMenu = new QMenu(tr("&View"), this);
	viewMenu->addAction(m_updateViewsAct);
	viewMenu->addSeparator();
	viewMenu->addAction(videoResAct);

	QAction *helpAct = new QAction(tr("Help"), this);
	connect(helpAct, SIGNAL(triggered()), this, SLOT(help()));
	QMenu *helpMenu = new QMenu(tr("&Help"), this);
	helpMenu->addAction(helpAct);

	menuBar()->addMenu(fileMenu);
	menuBar()->addMenu(viewMenu);
	menuBar()->addMenu(helpMenu);

	m_toolWindow = new QWidget();
	m_toolWindow->setWindowTitle("Tools");
	QVBoxLayout *layout = new QVBoxLayout();
	
	m_angleList = new QListWidget(m_toolWindow);
	connect(m_angleList, SIGNAL(itemClicked(QListWidgetItem *)), this, SLOT(itemClicked(QListWidgetItem *)));
	layout->addWidget(m_angleList);

	QPushButton *growFeature = new QPushButton("&Grow feature!", m_toolWindow);
	connect(growFeature, SIGNAL(clicked()), this, SLOT(growFeature()));
	
	m_startButton = new QPushButton("&Start!", m_toolWindow);

	connect(m_startButton, SIGNAL(clicked()), this, SLOT(go()));
	layout->addWidget(growFeature);
	layout->addWidget(m_startButton);
	
	m_toolWindow->setLayout(layout);

	m_fmt.setSampleBuffers(true);

	m_dataModel = new DataModel();
	m_widget = new GLWidget(m_fmt, this, m_dataModel);
	setCentralWidget(m_widget);
	
	m_running = false;
	m_timerOffset = 0;
	m_boldFont.setBold(true);

	resize(1000, 850);
	m_toolWindow->resize(275, 575);
	move(400, 0);

	connect(m_dataModel, SIGNAL(modelChangeEvent(DataModel::EChangeLevel)), 
		this, SLOT(modelChangeEvent(DataModel::EChangeLevel)));
	
	m_dataModel->load(params);
	m_toolWindow->show();
}

GLViewer::~GLViewer() {
	delete m_dataModel;
}

void GLViewer::loadMDLComparison() {
	QString fileName = QFileDialog::getOpenFileName(this, tr("Load curves from MDL .."), 
		QDir::currentPath(), "MDL files (*.mdl)");
	if (!fileName.isEmpty()) 
		m_dataModel->loadMDL(fileName.toStdString(), true);
}

void GLViewer::loadMDL() {
	QString fileName = QFileDialog::getOpenFileName(this, tr("Load curves from MDL .."), 
		QDir::currentPath(), "MDL files (*.mdl)");
	if (!fileName.isEmpty()) 
		m_dataModel->loadMDL(fileName.toStdString(), false);
}

void GLViewer::saveMDL() {
	QString fileName = QFileDialog::getSaveFileName(this, tr("Save curves to MDL .."), 
		QDir::currentPath(), "MDL files (*.mdl)");
	if (!fileName.isEmpty()) 
		m_dataModel->saveMDL(fileName.toStdString(), m_dataModel->getHairFibers());
}

void GLViewer::saveOBJ() {
	QString fileName = QFileDialog::getSaveFileName(this, tr("Save curves to OBJ .."), 
		QDir::currentPath(), "OBJ files (*.obj)");
	if (!fileName.isEmpty()) 
		m_dataModel->saveOBJ(fileName.toStdString());
}

void GLViewer::closeEvent(QCloseEvent *) {
	QApplication::exit(0);
}

void GLViewer::help() {
	QMessageBox::information(this, "Help", "Controls:\n"
		"- LMB controls the trackball rotation\n"
		"- RMB and the mouse wheel both control zoom\n"
		"- Shift-LMB translates around", QMessageBox::Ok);
}

void GLViewer::growFeature() {
	m_dataModel->getThread(0)->grow(true);
}

void GLViewer::clearFibers() {
	m_dataModel->clearFibers();
}

void GLViewer::go() {
	m_running = !m_running;
	if (m_running) {
		m_startButton->setText("&Stop!");
	} else {
		m_startButton->setText("&Start!");
	}
	m_dataModel->go();
}

void GLViewer::modelChangeEvent(DataModel::EChangeLevel level) {
	const Parameters &params = m_dataModel->getParameters();
	if (level == DataModel::NewFile) {
		m_angleList->clear();
		for (int i=0; i<params.numAngles; ++i) {
			QListWidgetItem* item = new QListWidgetItem();
			item->setText(formatString("Angle %i", i).c_str());
			item->setFlags(Qt::ItemIsUserCheckable | Qt::ItemIsEnabled);
			if (m_dataModel->getAngleVisibility(i))
				item->setCheckState(Qt::Checked);
			else
				item->setCheckState(Qt::Unchecked);
			m_angleList->addItem(item);
		}
	}
}

void GLViewer::itemClicked(QListWidgetItem *item) {
	m_dataModel->setAngleVisibility(m_angleList->row(item),
		item->checkState() == Qt::Checked);
}

void GLViewer::updateViews() {
	m_dataModel->setUpdateViews(m_updateViewsAct->isChecked());
}

void GLViewer::videoResolution() {
	m_widget->setMaximumSize(QSize(854, 480));
	m_widget->setMinimumSize(QSize(854, 480));
	resize(QSize(854, 480));
}

int main(int argc, char **argv) {
	Parameters parameters(argc, argv);
	QApplication app(argc, argv);
	cout << parameters.toString() << endl;
	try {
		if (parameters.filter) {
			runFilterStage(parameters);
		} else {
			qRegisterMetaType<DataModel::EChangeLevel>("DataModel::EChangeLevel");
			QThread::currentThread()->setPriority(QThread::HighestPriority);
			GLViewer glViewer(parameters);
			glViewer.show();
			return app.exec();
		}
	} catch (const std::exception &e) {
		cout << "Caught an exception: " << e.what() << endl;
	}
	return 0;
}

Q_DECLARE_METATYPE(DataModel::EChangeLevel);

