#include <hair/glwidget.h>
#include <hair/trackball.h>
#include <hair/wthread.h>

GLWidget::GLWidget(const QGLFormat &fmt, QWidget *parent, DataModel *model) : QGLWidget(fmt, parent), m_dataModel(model) {
	m_dist = 0;
	setFocusPolicy(Qt::StrongFocus);
	connect(model, SIGNAL(modelChangeEvent(DataModel::EChangeLevel)), 
		this, SLOT(modelChangeEvent(DataModel::EChangeLevel)), Qt::AutoConnection);

	srand(0);
	for (int i=0; i<MAX_ACTIVESET_SIZE; ++i) {
		float R = rand() / (float) RAND_MAX;
		float G = rand() / (float) RAND_MAX;
		float B = rand() / (float) RAND_MAX;
		float mult = 1.0 / std::max(std::max(R, G), B);
		m_colors[i][0] = R*mult;
		m_colors[i][1] = G*mult;
		m_colors[i][2] = B*mult;
	}
}

void GLWidget::modelChangeEvent(DataModel::EChangeLevel level) {
	switch (level) {
		case DataModel::NewFile: {
			/* Completely reset the view */
			m_mouseButton = 0;
			trackball(m_quat, 0.0, 0.0, 0.0, 0.25);
			m_offset[0] = m_offset[1] = m_offset[2] = 0;
			m_offset[1] = -100;
			m_dist = 4000;
			m_bounds = OBB(m_dataModel->getAABB());
			m_displayBounds = m_bounds;
			m_displayBounds.m_extents.y *= .005f;
			m_rotationY = 0;
		}
		case DataModel::FocusAndRefresh:
			activateWindow();
		case DataModel::Refresh:
		case DataModel::Repaint:
			if (m_dataModel->getUpdateViews())
				updateGL();
	}
}
	
QSize GLWidget::minimumSizeHint() const {
	return QSize(50, 50);
}

QSize GLWidget::sizeHint() const {
	return QSize(400, 400);
}

void GLWidget::initializeGL() {
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
}

void glSolidCube(const Point3 &p, GLdouble size) {
#   define V(a,b,c) glVertex3d( p.x a size, p.y b size, p.z c size );
#   define N(a,b,c) glNormal3d( a, b, c );
    glBegin(GL_QUADS);
        N( 1.0, 0.0, 0.0); V(+,-,+); V(+,-,-); V(+,+,-); V(+,+,+);
        N( 0.0, 1.0, 0.0); V(+,+,+); V(+,+,-); V(-,+,-); V(-,+,+);
        N( 0.0, 0.0, 1.0); V(+,+,+); V(-,+,+); V(-,-,+); V(+,-,+);
        N(-1.0, 0.0, 0.0); V(-,-,+); V(-,+,+); V(-,+,-); V(-,-,-);
        N( 0.0,-1.0, 0.0); V(-,-,+); V(-,-,-); V(+,-,-); V(+,-,+);
        N( 0.0, 0.0,-1.0); V(-,-,-); V(-,+,-); V(+,+,-); V(+,-,-);
    glEnd();
#   undef V
#   undef N
}

void glSolidCube(const Point3 &p, Vector3 size) {
#   define V(a,b,c) glVertex3d( p.x a size.x, p.y b size.y, p.z c size.z );
#   define N(a,b,c) glNormal3d( a, b, c );
    glBegin(GL_QUADS);
        N( 1.0, 0.0, 0.0); V(+,-,+); V(+,-,-); V(+,+,-); V(+,+,+);
        N( 0.0, 1.0, 0.0); V(+,+,+); V(+,+,-); V(-,+,-); V(-,+,+);
        N( 0.0, 0.0, 1.0); V(+,+,+); V(-,+,+); V(-,-,+); V(+,-,+);
        N(-1.0, 0.0, 0.0); V(-,-,+); V(-,+,+); V(-,+,-); V(-,-,-);
        N( 0.0,-1.0, 0.0); V(-,-,+); V(-,-,-); V(+,-,-); V(+,-,+);
        N( 0.0, 0.0,-1.0); V(-,-,-); V(-,+,-); V(+,+,-); V(+,-,-);
    glEnd();
#   undef V
#   undef N
}

/* http://www.tecgraf.puc-rio.br/~mgattass/color/HSVtoRGB.htm */ 
static void HSVtoRGB(double h, double s, double v, double* r, double* g, double* b) {
	if (s == 0) {
		*r = v;
		*g = v;
		*b = v;
	} else {
		double var_h = h * 6;
		double var_i = floor(var_h);
		double var_1 = v * (1 - s);
		double var_2 = v * (1 - s * (var_h - var_i));
		double var_3 = v * (1 - s * (1 - (var_h - var_i)));
		if      (var_i == 0) { *r = v     ; *g = var_3 ; *b = var_1; }
		else if (var_i == 1) { *r = var_2 ; *g = v     ; *b = var_1; }
		else if (var_i == 2) { *r = var_1 ; *g = v     ; *b = var_3; }
		else if (var_i == 3) { *r = var_1 ; *g = var_2 ; *b = v;     }
		else if (var_i == 4) { *r = var_3 ; *g = var_1 ; *b = v;     }
		else                 { *r = v     ; *g = var_1 ; *b = var_2; }
	}
}

void GLWidget::paintGL() {
	const Parameters &params = m_dataModel->getParameters();

	if (!m_dist)
		return;

	glDisable(GL_SCISSOR_TEST);
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_SCISSOR_TEST);
	glDepthFunc(GL_LEQUAL);

	GLfloat mat[4][4];
	build_rotmatrix(mat, m_quat);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	glViewport(0, 0, width(), height());
	glScissor(0, 0, width(), height());

	gluPerspective(45.0f, width() / (float) height(), 10.0f, 1e5);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(0, 0, m_dist, 0, 0, 0, 0, -1, 0);

	glMultMatrixf(&mat[0][0]);
	glTranslatef(m_offset[0], m_offset[1], m_offset[2]);
	glScalef(-1, 1, -1);
	glRotatef(m_rotationY, 0, 1, 0);

	glColor4f(.4f, .4f, .4f, 1.0f);
	m_bounds.drawGL();
	m_displayBounds.drawGL();

	for (int threadIdx=0; threadIdx < m_dataModel->getThreadCount(); ++threadIdx) {
		AABB seedAABB = m_dataModel->getThread(threadIdx)->getSeedAABB();
		if (seedAABB.isValid())
			OBB(seedAABB).drawGL();
	}

	for (int activeSetIdx=0; activeSetIdx < m_dataModel->getActiveSetCount(); ++activeSetIdx) {
		const ActiveSet &activeSet = m_dataModel->getActiveSet(activeSetIdx);
	
		activeSet.lock();
		const HairFiber &fiber = activeSet.fiber;
		for (size_t i=0; i<fiber.vertices.size(); ++i) {

			glColor4f(0, 0, .5, 1);
			glBegin(GL_LINE_STRIP);
			for (size_t i=0; i<fiber.vertices.size(); ++i) {
				Point3 p = fiber.vertices[i];
				glVertex3f(p.x, p.y, p.z);
			}
			glEnd();
		}

		if (activeSet.nActive > 0) {
			Vector3 point = activeSet.p + activeSet.d*GROW_STEPSIZE;
			for (int angle=0; angle < params.numAngles; ++angle) {	
				glPushMatrix();
				glRotatef((360.0/params.numAngles)*angle, 0.0f, 1.0f, 0.0f);
				glScalef(1,1,-1.0/params.aspectZ);
				glTranslatef(-params.centerX, -params.yres/2.0, -params.centerDepth);
				glColor4f(m_colors[angle][0], m_colors[angle][1], m_colors[angle][2], 1.0f);

#if 0
				/* Uncomment to display the feature data during the growing process */

				for (std::set<KDTree::Observation *>::iterator it = activeSet.observations[angle].begin();
					it != activeSet.observations[angle].end(); ++it) {
					const Point3s &p = (*it)->pos;
					glSolidCube(Point3(p.x, p.y, p.z), .5);
				}
#endif
				glPopMatrix();

				if (!activeSet.active[angle])
					continue;
				
				const Vector3 &sheetSize = m_dataModel->getParameters().sheetSize;

				Vector3 dxdz = m_dataModel->dXYdZ_vec(point, angle);
				dxdz = normalize(dxdz - activeSet.d*dot(dxdz, activeSet.d));
				Vector3 d = activeSet.d;
				Vector3 d3 = cross(dxdz, d);

				glColor4f(m_colors[angle][0], m_colors[angle][1], m_colors[angle][2], 1.0f);
	
				GLfloat mat[4][4];
				mat[0][0] = dxdz.x; mat[0][1] = dxdz.y; mat[0][2] = dxdz.z; mat[0][3] = 0;
				mat[1][0] = d.x;    mat[1][1] = d.y;    mat[1][2] = d.z;    mat[1][3] = 0;
				mat[2][0] = d3.x;   mat[2][1] = d3.y;   mat[2][2] = d3.z;   mat[2][3] = 0;
				mat[3][0] = 0;      mat[3][1] = 0;      mat[3][2] = 0;      mat[3][3] = 1;

				glPushMatrix();
				glTranslatef(point.x, point.y, point.z);
				glMultMatrixf(&mat[0][0]);
				glScalef(sheetSize.x, sheetSize.y, sheetSize.z);
				glSolidCube(Vector3(0,0,0), 1);
				glPopMatrix();
			}

			glBegin(GL_LINES);
			glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
			Vector3 d = activeSet.d * 50;
			glVertex3f(activeSet.p.x, activeSet.p.y, activeSet.p.z);
			glVertex3f(activeSet.p.x + d.x, activeSet.p.y + d.y, activeSet.p.z + d.z);
			glEnd();
		}

		activeSet.unlock();
	}
/*
	const GLdouble eqn1[] = { 0, -1, 0, m_displayBounds.getCenter().y + m_displayBounds.getExtents(1)};
	const GLdouble eqn2[] = { 0, 1, 0, -(m_displayBounds.getCenter().y - m_displayBounds.getExtents(1))};
	glClipPlane(GL_CLIP_PLANE0, eqn1);
	glEnable(GL_CLIP_PLANE0);
	glClipPlane(GL_CLIP_PLANE1, eqn2);
	glEnable(GL_CLIP_PLANE1);
*/
	const std::vector<HairFiber> &hairFibers = m_dataModel->getHairFibers();
	const std::vector<HairFiber> &comparisonHairFibers = m_dataModel->getComparisonHairFibers();

	m_dataModel->lock();
	for (size_t i=0; i<hairFibers.size(); ++i) {
		const HairFiber &fiber = hairFibers[i];

		double r, g, b;
		HSVtoRGB(228/360.0f, fiber.saturation, 0.64, &r, &g, &b);
		glColor4f(r, g, b, 1.0f);
		glBegin(GL_LINE_STRIP);
		for (size_t i=0; i<fiber.vertices.size(); ++i) {
			Point3 p = fiber.vertices[i];
			glVertex3f(p.x, p.y, p.z);
		}
		glEnd();
	}
	
	for (size_t i=0; i<comparisonHairFibers.size(); ++i) {
		const HairFiber &fiber = comparisonHairFibers[i];

		glBegin(GL_LINE_STRIP);
		glColor4f(61/255.0f, 189/255.0f, 61/255.0f, 1.0f);
		double r, g, b;
		HSVtoRGB(120/360.0f, fiber.saturation, 0.64, &r, &g, &b);
		glColor4f(r, g, b, 1.0f);
		for (size_t i=0; i<fiber.vertices.size(); ++i) {
			Point3 p = fiber.vertices[i];
			glVertex3f(p.x, p.y, p.z);
		}
		glEnd();
	}
	m_dataModel->unlock();

	for (int angle=0; angle < params.numAngles; ++angle) {
		if (!m_dataModel->getAngleVisibility(angle))
			continue;
		glPushMatrix();
		glRotatef((360.0/params.numAngles)*angle, 0.0f, 1.0f, 0.0f);
		glScalef(1,1,-1.0/params.aspectZ);
		glTranslatef(-params.centerX, -params.yres/2.0, -params.centerDepth);
		glColor4f(m_colors[angle][0], m_colors[angle][1], m_colors[angle][2], 1.0f);

		m_result.clear();
		m_dataModel->obbQuery(m_displayBounds, angle, m_result);

		for (size_t i=0; i<m_result.size(); ++i) {
			const Point3s &p = m_result[i]->pos;
			glSolidCube(Point3(p.x, p.y, p.z), .5);
		}
		glPopMatrix();
	}

/*
	glDisable(GL_CLIP_PLANE0);
	glDisable(GL_CLIP_PLANE1);
*/

	glBegin(GL_LINES);
	glColor4f(1.0f, 0.0f, 0.0f, 0.3f);
	glVertex3f(100, 0, 0);
	glVertex3f(0, 0, 0);
	glColor4f(0.0f, 1.0f, 0.0f, 0.3f);
	glVertex3f(0, 0, 100);
	glVertex3f(0, 0, 0);
	glColor4f(0.0f, 0.0f, 1.0f, 0.3f);
	glVertex3f(0, 100, 0);
	glVertex3f(0, 0, 0);
	glEnd();

	GLuint glError = glGetError();
	if (glError)
		throw std::runtime_error(formatString("GLWidget.paintGL - OpenGL Error: %s", gluErrorString(glError)));
}

   
void GLWidget::resizeGL(int width, int height) {
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, (float) width / (float) height, 0.1f, 1000.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void GLWidget::mousePressEvent(QMouseEvent *event) {
	m_lastPos = event->pos();
	m_mouseButton = event->button();
}

void GLWidget::mouseReleaseEvent(QMouseEvent *) {
	m_mouseButton = 0;
}

void GLWidget::wheelEvent(QWheelEvent *event) {
	float dz = event->delta() * 2.0f;

	if (event->orientation() == Qt::Vertical)
		dz *= -1;
	m_dist += dz;

	event->accept();
	updateGL();
}

void GLWidget::mouseMoveEvent(QMouseEvent *event) {
	int windowWidth = width(), windowHeight = height();
	if (m_mouseButton == Qt::LeftButton) {
		if ((event->modifiers() & Qt::ShiftModifier) != 0) {
			float dx = -(event->x() - m_lastPos.x()) * 2.0f;
			float dy = (event->y() - m_lastPos.y()) * 2.0f;
			GLfloat mat[4][4];
			build_rotmatrix(mat, m_quat);
			m_offset[0] += dx*mat[0][0] + dy*mat[0][1];
			m_offset[1] += dx*mat[1][0] + dy*mat[1][1];
			m_offset[2] += dx*mat[2][0] + dy*mat[2][1];
		} else {
			float q[4];

			trackball(q,  -(2.0 * m_lastPos.x() - windowWidth)  / windowWidth,
					(2.0 * m_lastPos.y() - windowHeight) / windowHeight,
					-(2.0 * event->x() - windowWidth)  / windowWidth,
					(2.0 * event->y() - windowHeight) / windowHeight);

			add_quats(q, m_quat, m_quat);
		}
		event->accept();
		updateGL();
	} else if (m_mouseButton == Qt::RightButton) {
		float dy = -(event->y() - m_lastPos.y()) * 2.0f;
		m_dist -= dy;
		event->accept();
		updateGL();
	}

	m_lastPos = event->pos();
}

QImage GLWidget::screenshot() {
	QPixmap pixmap=renderPixmap(width(), height());
	return pixmap.toImage();
}

void GLWidget::setRotationY(Float amount) {
	m_rotationY = amount;
}

void GLWidget::keyPressEvent(QKeyEvent *event) {
	const Parameters &params = m_dataModel->getParameters();
	switch (event->key()) {
		case Qt::Key_Escape:
			exit(0);
			break;
		case Qt::Key_W:
			m_dataModel->getThread(0)->getSeedAABB().min.z += 30;
			m_dataModel->getThread(0)->getSeedAABB().max.z += 30;
			cout << m_dataModel->getThread(0)->getSeedAABB().toString() << endl;
			event->accept();
			break;
		case Qt::Key_S:
			m_dataModel->getThread(0)->getSeedAABB().min.z -= 30;
			m_dataModel->getThread(0)->getSeedAABB().max.z -= 30;
			cout << m_dataModel->getThread(0)->getSeedAABB().toString() << endl;
			event->accept();
			break;
		case Qt::Key_A:
			m_dataModel->getThread(0)->getSeedAABB().min.x -= 30;
			m_dataModel->getThread(0)->getSeedAABB().max.x -= 30;
			cout << m_dataModel->getThread(0)->getSeedAABB().toString() << endl;
			event->accept();
			break;
		case Qt::Key_D:
			m_dataModel->getThread(0)->getSeedAABB().min.x += 30;
			m_dataModel->getThread(0)->getSeedAABB().max.x += 30;
			cout << m_dataModel->getThread(0)->getSeedAABB().toString() << endl;
			event->accept();
			break;
		case Qt::Key_R:
			m_dataModel->getThread(0)->getSeedAABB().min.y -= 30;
			m_dataModel->getThread(0)->getSeedAABB().max.y -= 30;
			cout << m_dataModel->getThread(0)->getSeedAABB().toString() << endl;
			event->accept();
			break;
		case Qt::Key_F:
			m_dataModel->getThread(0)->getSeedAABB().min.y += 30;
			m_dataModel->getThread(0)->getSeedAABB().max.y += 30;
			cout << m_dataModel->getThread(0)->getSeedAABB().toString() << endl;
			event->accept();
			break;
		case Qt::Key_BracketLeft:
			m_displayBounds.m_extents.y /= 2;
			if (m_displayBounds.m_extents.y < 1)
				m_displayBounds.m_extents.y = 1;
			event->accept();
			break;
		case Qt::Key_BracketRight:
			m_displayBounds.m_extents.y *= 2;
			if (m_displayBounds.m_extents.y > params.yres/2)
				m_displayBounds.m_extents.y = params.yres/2;
			event->accept();
			break;
		case Qt::Key_Up:
			m_displayBounds.m_center.y -= m_displayBounds.m_extents.y;
			event->accept();
			break;
		case Qt::Key_Down:
			m_displayBounds.m_center.y += m_displayBounds.m_extents.y;
			event->accept();
			break;
	}

	if (m_displayBounds.m_center.y + m_displayBounds.m_extents.y > params.yres/2)
		m_displayBounds.m_center.y = params.yres/2 - m_displayBounds.m_extents.y;
	if (m_displayBounds.m_center.y - m_displayBounds.m_extents.y < -params.yres/2)
		m_displayBounds.m_center.y = -params.yres/2  + m_displayBounds.m_extents.y;

	updateGL();
}

