#include <hair/imageio.h>
#include <png.h>

/* Endianness conversion */
#ifndef WIN32
#include <arpa/inet.h>
#endif

/* Load an 8-bit PNG file */
IplImage *cvLoadPNG8(const std::string &fname) {
	cout << "    + Loading \"" << fname << "\"" << endl;
	FILE *file = fopen(fname.c_str(), "rb");
	if (file == NULL) {
		cerr << "Error: Could not open \"" << fname << "\"!" << endl;
		exit(-1);
	}

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	png_infop info = png_create_info_struct(png_ptr);
	if (info == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	if (setjmp(png_ptr->jmpbuf)) {
		cerr << "Error: \"" << fname << "\" could not be read - broken file?" 
			 << endl;
		exit(-1);
		
	}

	png_init_io (png_ptr, file);
	png_read_info(png_ptr, info);
	int width = info->width, height = info->height, channels = info->channels;

	if (info->bit_depth != 8) {
		cout << "Expected a 8-bit image!" << endl;
		exit(-1);
	}

	png_bytepp rows = new png_bytep[height];
	png_bytep data = new png_byte[width*height*channels];

	for (int i=0; i<height; i++)
		rows[i] = data + width*channels*i;

	png_read_image(png_ptr, (png_bytepp) rows);
	png_read_end(png_ptr, info);
	png_destroy_read_struct(&png_ptr, &info, NULL);

	IplImage *image = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, channels);

	for (int y=0; y<height; ++y) {
		for (int x=0; x<width; ++x) {
			for (int c=0; c<channels; ++c) {
				((png_byte *) image->imageData + image->widthStep*y + x*channels)[c]
					= (data + width*channels*y + x*channels)[c];
			}
		}
	}

	delete[] data;
	delete[] rows;

	fclose(file);
	return image;
}

/* Load a 16-bit PNG file */
IplImage *cvLoadPNG16(const std::string &fname) {
	cout << "    + Loading \"" << fname << "\"" << endl;
	FILE *file = fopen(fname.c_str(), "rb");
	if (file == NULL) {
		cerr << "Error: Could not open \"" << fname << "\"!" << endl;
		exit(-1);
	}

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	png_infop info = png_create_info_struct(png_ptr);
	if (info == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	if (setjmp(png_ptr->jmpbuf)) {
		cerr << "Error: \"" << fname << "\" could not be read - broken file?" 
			 << endl;
		exit(-1);
		
	}

	png_init_io (png_ptr, file);
	png_read_info(png_ptr, info);
	int width = info->width, height = info->height, channels = info->channels;

	if (info->bit_depth != 16) {
		cout << "Expected a 16-bit image!" << endl;
		exit(-1);
	}

	png_uint_16pp rows = new png_uint_16p[height];
	png_uint_16p data = new png_uint_16[width*height*channels];

	for (int i=0; i<height; i++)
		rows[i] = data + width*channels*i;

	png_read_image(png_ptr, (png_bytepp) rows);
	png_read_end(png_ptr, info);
	png_destroy_read_struct(&png_ptr, &info, NULL);

	IplImage *image = cvCreateImage(cvSize(width, height), IPL_DEPTH_16S, channels);
	for (int y=0; y<height; ++y) {
		for (int x=0; x<width; ++x) {
			for (int c=0; c<channels; ++c) {
				((short *) (image->imageData + image->widthStep*y) + x*channels)[c]
					= htons((data + width*channels*y + x*channels)[c]);
			}
		}
	}

	delete[] data;
	delete[] rows;

	fclose(file);
	return image;
}

/* Save an 8-bit PNG file */
void cvSavePNG8(const std::string &fname, IplImage *image, CvRect roi) {
	cout << "    + Writing \"" << fname << "\"" << endl;

	if (roi.x == -1 && roi.y == -1 && roi.width == -1 && roi.height == -1) {
		roi.x = 0;
		roi.y = 0;
		roi.width = image->width;
		roi.height = image->height;
	}

	if (image->depth != (int) IPL_DEPTH_8U) {
		cerr << "Expected an 8-bit image!" << endl;
		exit(-1);
	}
	
	if (image->nChannels != 1 && image->nChannels != 3) {
		cerr << "Expected a 1- or 3 channel image!" << endl;
		exit(-1);
	}

	FILE *file = fopen(fname.c_str(), "wb");
	if (file == NULL) {
		cerr << "Error: Could not open \"" << fname << "\"!" << endl;
		exit(-1);
	}

	png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	png_infop info = png_create_info_struct(png_ptr);
	if (info == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	if (setjmp(png_ptr->jmpbuf)) {
		cerr << "Error: \"" << fname << "\" could not be written!" 
			 << endl;
		exit(-1);
		
	}

	png_init_io(png_ptr, file);
	png_set_compression_level(png_ptr, 3);
	png_set_IHDR(png_ptr, info, roi.width, roi.height, 8,
		(image->nChannels == 1) ? PNG_COLOR_TYPE_GRAY : PNG_COLOR_TYPE_RGB,
		PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
	png_write_info(png_ptr, info);

	png_bytepp rows = new png_bytep[roi.height];
	png_bytep data = new png_byte[roi.width*roi.height*image->nChannels];

	for (int i=0; i<roi.height; i++)
		rows[i] = data + roi.width*image->nChannels*i;

	for (int y=0; y<roi.height; ++y) {
		for (int x=0; x<roi.width; ++x) {
			for (int c=0; c<image->nChannels; ++c) {
				(data + roi.width*image->nChannels*y + x*image->nChannels)[c] =
					((png_byte *)(image->imageData + image->widthStep*(y+roi.y)) + (x+roi.x)*image->nChannels)[c];
			}
		}
	}

	png_write_image(png_ptr, (png_bytepp) rows);
	png_write_end(png_ptr, info);
	png_destroy_write_struct(&png_ptr, &info);

	delete[] data;
	delete[] rows;
	fclose(file);
}


/* Save a 16-bit PNG file */
void cvSavePNG16(const std::string &fname, IplImage *image, CvRect roi) {
	cout << "    + Writing \"" << fname << "\"" << endl;	

	if (image->depth != (int) IPL_DEPTH_16S) {
		cerr << "Expected a 16-bit image!" << endl;
		exit(-1);
	}
	
	if (image->nChannels != 1 && image->nChannels != 3) {
		cerr << "Expected a 1- or 3 channel image!" << endl;
		exit(-1);
	}

	if (roi.x == -1 && roi.y == -1 && roi.width == -1 && roi.height == -1) {
		roi.x = 0;
		roi.y = 0;
		roi.width = image->width;
		roi.height = image->height;
	}

	FILE *file = fopen(fname.c_str(), "wb");
	if (file == NULL) {
		cerr << "Error: Could not open \"" << fname << "\"!" << endl;
		exit(-1);
	}
	png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	png_infop info = png_create_info_struct(png_ptr);
	if (info == NULL) {
		cout << "Error: could not initialize libpng!" << endl;
		exit(-1);
	}

	if (setjmp(png_ptr->jmpbuf)) {
		cerr << "Error: \"" << fname << "\" could not be written!" 
			 << endl;
		exit(-1);
		
	}

	png_init_io(png_ptr, file);
	png_set_compression_level(png_ptr, 3);
	png_set_IHDR(png_ptr, info, roi.width, roi.height, 16,
		(image->nChannels == 1) ? PNG_COLOR_TYPE_GRAY : PNG_COLOR_TYPE_RGB,
		PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
	png_write_info(png_ptr, info);

	png_uint_16pp rows = new png_uint_16p[roi.height];
	png_uint_16p data = new png_uint_16[roi.width*roi.height*image->nChannels];

	for (int i=0; i<roi.height; i++)
		rows[i] = data + roi.width*image->nChannels*i;
	
	for (int y=0; y<roi.height; ++y) {
		for (int x=0; x<roi.width; ++x) {
			for (int c=0; c<image->nChannels; ++c) {
				(data + roi.width*image->nChannels*y + x*image->nChannels)[c] =
					htons(((short *)(image->imageData + image->widthStep*(y + roi.y)) + (x+roi.x)*image->nChannels)[c]);
			}
		}
	}

	png_write_image(png_ptr, (png_bytepp) rows);
	png_write_end(png_ptr, info);
	png_destroy_write_struct(&png_ptr, &info);

	delete[] data;
	delete[] rows;
	fclose(file);
}

uint8_t cvGetInterpolated(const IplImage *src, float y, float x) {
	int xPos = (int) x, yPos = (int) y;
	float alpha = x-xPos, beta = y-yPos;
	char *data = src->imageData + src->widthStep*yPos 
		+ sizeof(uint8_t) * xPos*src->nChannels;

	int yJump = 1;
	if (yPos+1>= src->height)
		yJump -= src->height;

	const uint8_t 
		A = *((uint8_t *) data),
		B = (xPos+1 < src->width)  ? *((uint8_t *) data + src->nChannels) : 0,
		C = *((uint8_t *) (data + yJump*src->widthStep)),
		D = (xPos+1 < src->width) ? *((uint8_t *) (data + yJump*src->widthStep) + src->nChannels) : 0;

	return (uint8_t) 
	      (A * ((1-alpha) * (1-beta))
		 + B * (   alpha  * (1-beta))
		 + C * ((1-alpha) *    beta)
		 + D *     alpha  *    beta);
}

short cvGetInterpolated16(const IplImage *src, float y, float x) {
	int xPos = (int) x, yPos = (int) y;
	float alpha = x-xPos, beta = y-yPos;
	char *data = src->imageData + src->widthStep*yPos 
		+ sizeof(short) * xPos*src->nChannels;

	int yJump = 1;
	if (yPos+1>= src->height)
		yJump -= src->height;

	const short 
		A = *((short *) data),
		B = (xPos+1 < src->width)  ? *((short *) data + src->nChannels) : 0,
		C = *((short *) (data + yJump*src->widthStep)),
		D = (xPos+1 < src->width) ? *((short *) (data + yJump*src->widthStep) + src->nChannels) : 0;

	return (short) 
	      (A * ((1-alpha) * (1-beta))
		 + B * (   alpha  * (1-beta))
		 + C * ((1-alpha) *    beta)
		 + D *     alpha  *    beta);
}

