#include <hair/kdtree.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>

KDTree::KDTree(const std::string &filename, const Vector3 &scale, bool create) 
	:  m_observations(NULL), m_scale(scale), m_balanced(!create), m_filename(filename), m_file(NULL), m_fd(-1) {
	if (create) {
		m_file = fopen(filename.c_str(), "wb");
		if (!m_file) 
			throw std::runtime_error(formatString("Could not create %s!", filename.c_str()));
		if (fseek(m_file, sizeof(AABB3s) + sizeof(size_t) + sizeof(Observation), SEEK_SET))
			throw std::runtime_error(formatString("Could not seek in %s!", filename.c_str()));
		m_observationCount = 0;
	} else {
		m_fd = open(filename.c_str(), O_RDWR);
		struct stat stbuf;
		if (m_fd == -1)
			throw std::runtime_error(formatString("Could not open %s!", filename.c_str()));
		if (read(m_fd, &m_aabb, sizeof(AABB3s)) != sizeof(AABB3s))
			throw std::runtime_error(formatString("Could not load %s!", filename.c_str()));
		if (read(m_fd, &m_observationCount, sizeof(size_t)) != sizeof(size_t))
			throw std::runtime_error(formatString("Could not load %s!", filename.c_str()));
		fstat(m_fd, &stbuf);
	
		m_lastInnerNode = m_observationCount/2;
		m_lastRChildNode = (m_observationCount-1)/2;

		m_mmapSize = stbuf.st_size;
		m_mmapPtr = mmap(NULL, m_mmapSize, PROT_READ | PROT_WRITE, MAP_SHARED, m_fd, 0);
		if (m_mmapPtr == MAP_FAILED)
			throw std::runtime_error(formatString("Could not map %s to memory!", filename.c_str()));
		m_observations = (Observation *) ((char *) m_mmapPtr + sizeof(AABB3s) + sizeof(size_t));
	}
}

bool KDTree::storeObservations(const std::vector<Observation> &observations) {
	assert(!m_balanced);
	if (fwrite(&observations[0], sizeof(Observation), observations.size(), m_file) != observations.size())
		throw std::runtime_error("Error while writing observations to disk");
	for (size_t i=0; i<observations.size(); ++i)
		m_aabb.expandBy(observations[i].pos);
	m_observationCount += observations.size();
	return true;
}


KDTree::~KDTree() {
	if (m_file)
		fclose(m_file);
	if (m_fd != -1) {
		munmap(m_mmapPtr, m_mmapSize);
		close(m_fd);
	}
}

/**
 * Relaxed partitioning algorithm based on code in stl_algo.h and Jensen's
 * reference implementation. This is *much* faster than std::partition
 * by relaxing the partitioning rules, making use of random access iterators 
 * and assuming that there is a guard element following the right boundary. 
 *
 * This function accepts *two* predicates and assumes that 
 * pred1(x) == !pred2(x) for all given elements except for a set S, 
 * where both predicates return false. After the algorithm finishes,
 * elements in S may have been placed on either side. However, all elements
 * for which pred1(x) is true are guaranteed to be on the left side, and
 * all elements for which pred2(x) is true will have been moved to the
 * right side.
 * 
 * Having two predicates is important for an efficient kd-tree
 * balancing implementation. If pred2 is simply the opposite of pred1,
 * the balancing routine will slow down to a crawl when it is applied to
 * an array containing lots of elements with the same value as the pivot.
 * (In the context of photon mapping, this can for example happen when the 
 * scene contains an axis-aligned surface). Since it does not matter on 
 * which side of the split these elements end up, the relaxed method 
 * enables a much better separation into two sets.
 */
template<typename _RandomAccessIterator, 
	typename _Predicate1, typename _Predicate2>
	inline _RandomAccessIterator guarded_partition(
		_RandomAccessIterator start, _RandomAccessIterator end, 
		_Predicate1 pred1, _Predicate2 pred2) {
	end--;
	while (true) {
		while (pred1(*start)) /* Guarded */
			start++;
		while (pred2(*end) && end > start)
			end--;
		if (start >= end)
			break;
		std::iter_swap(start++, end--);
	}
	return start;
}

/**
 * This algorithm works similarly to QUICKSORT, although it does not sort
 * the array. Instead, it creates a partition such that there is an ordering 
 * of all entries with respect to the value at 'pivotIndex'. It does this by
 * repetetively partitioning according to some arbitrarily chosen pivot 
 * (currently the rightmost one) and then recursing either into the left or 
 * right halves.
 */
void KDTree::quickPartition(observation_iterator left, observation_iterator right, 
	observation_iterator pivot, int axis) const {
	right--;

	while (right > left) {
		const float pivotValue = (*right)->pos[axis];

		/* Relaxed quicksort-style partitioning routine. (See above
		   for an explanation). */
		observation_iterator mid = guarded_partition(left, right,
			functorLess(axis, pivotValue),
			functorGreater(axis, pivotValue)
		);

		/* Move the pivot in between the two sets */
		std::swap(*mid, *right);

		/* Is the requested pivot index located inside the left half? */
		if (mid > pivot)
			right = mid - 1;
		/* Is the requested pivot index located inside the right half? */
		else if (mid < pivot)
			left = mid + 1;
		else
			return;
	}
}

/**
 * This algorithm is based on Donald Knuth's book
 * "The Art of Computer Programming, Volume 3: Sorting and Searching"
 * (1st edition, section 5.2, page 595)
 *
 * Given a permutation and an array of values, it applies the permutation
 * in linear time without requiring additional memory. This is based on
 * the fact that each permutation can be decomposed into a disjoint set
 * of permutations, which can then be applied individually.
 */
template <typename T> void permute_inplace(T *values, std::vector<size_t> &perm) {
	for (size_t i=0; i<perm.size(); i++) {
		if (perm[i] != i) {
			/* The start of a new cycle has been found. Save
			   the value at this position, since it will be
			   overwritten */
			size_t j = i;
			T curval = values[i];

			do {
				/* Shuffle backwards */
				size_t k = perm[j];
				values[j] = values[k];

				/* Also fix the permutations on the way */
				perm[j] = j;
				j = k;

				/* Until the end of the cycle has been found */
			} while (perm[j] != i);

			/* Fix the final position with the saved value */
			values[j] = curval;
			perm[j] = j;
		}
	}
}

/**
 * Given a number of entries, this method calculates the maximum amount of
 * nodes on the left subtree of a left-balanced tree. There are two main 
 * cases here:
 * 
 * 1) It is possible to completely fill the left subtree
 * 2) It doesn't work - the last level contains too few nodes, e.g :
 *         O
 *        / \
 *       O   O
 *      /
 *     O
 * 
 * The function assumes that "count" > 1.
 */
size_t KDTree::leftSubtreeSize(size_t treeSize) const {
	/* Layer 0 contains one node */
	size_t p = 1;

	/* Traverse downwards until the first incompletely
	   filled tree level is encountered */
	while (2*p <= treeSize)
		p *= 2;

	/* Calculate the number of filled slots in the last level */
	size_t remaining = treeSize - p + 1;

	if (2*remaining < p) {
		/* Case 2: The last level contains too few nodes. Remove
		   overestimate from the left subtree node count and add
		   the remaining nodes */
		p = (p >> 1) + remaining;
	}

	return p - 1;
}

void KDTree::balance() {
	assert(!m_balanced);

	/* Update the file header */
	if (fseek(m_file, 0, SEEK_SET))
		throw std::runtime_error(formatString("Could not seek in %s!", m_filename.c_str()));
	if (fwrite(&m_aabb, sizeof(AABB3s), 1, m_file) != 1)
		throw std::runtime_error(formatString("Could not fwrite to %s!", m_filename.c_str()));
	if (fwrite(&m_observationCount, sizeof(size_t), 1, m_file) != 1)
		throw std::runtime_error(formatString("Could not fwrite to %s!", m_filename.c_str()));
	fclose(m_file);
	m_file = NULL;
	
	m_fd = open(m_filename.c_str(), O_RDWR);
	struct stat stbuf;
	if (m_fd == -1)
		throw std::runtime_error(formatString("Could not open %s!", m_filename.c_str()));
	fstat(m_fd, &stbuf);

	m_mmapSize = stbuf.st_size;
	m_mmapPtr = mmap(NULL, m_mmapSize, PROT_READ | PROT_WRITE, MAP_SHARED, m_fd, 0);
	if (m_mmapPtr == MAP_FAILED)
		throw std::runtime_error(formatString("Could not map %s to memory!", m_filename.c_str()));
	m_observations = (Observation *) ((char *) m_mmapPtr + sizeof(AABB3s) + sizeof(size_t));

	/* Shuffle pointers instead of copying observations back and forth */
	std::vector<observation_ptr> observationPointers(m_observationCount + 1);
	/* Destination for the final heap permutation. Indexed starting at 1 */
	std::vector<size_t> heapPermutation(m_observationCount + 1);

	heapPermutation[0] = 0;
	for (size_t i=0; i<=m_observationCount; i++)
		observationPointers[i] = &m_observations[i];

	cout << " - kd-tree: balancing " << m_observationCount << " observations .." << endl;
	balanceRecursive(observationPointers.begin(), observationPointers.begin()+1, 
		observationPointers.end(), heapPermutation, m_aabb, 1);

	/* 'heapPointers' now contains a permutation representing
	    the properly left-balanced kd-tree. Apply this permutation
		to the observation array. */
	permute_inplace(m_observations, heapPermutation);

	/* We want to quickly be able to determine whether a node at
	   a given index is an inner node (e.g. it has left or right
	   children) and specifically, if it has a right child */
	m_lastInnerNode = m_observationCount/2;
	m_lastRChildNode = (m_observationCount-1)/2;
	m_balanced = true;
}

void KDTree::balanceRecursive(observation_iterator basePtr, 
	observation_iterator sortStart,
	observation_iterator sortEnd, 
	std::vector<size_t> &heapPermutation,
	AABB3s &aabb, size_t heapIndex) const {

	/* A fully left-balanced binary tree has this many nodes on its
	   left subtree */
	size_t leftSize = leftSubtreeSize(sortEnd - sortStart);

	/* We choose a pivot index such that the resulting tree satisfies
       this property */
	observation_iterator pivot = sortStart + leftSize;

	/* Splitting along the axis with the widest spread
       works well in practice and is cheap to compute (p.71) */

	Vector3 d(
		aabb.max.x - aabb.min.x,
		aabb.max.y - aabb.min.y,
		aabb.max.z - aabb.min.z
	);

	int splitAxis;
	if (d.x > d.y && d.x > d.z)
		splitAxis = 0;
	else if (d.y > d.z)
		splitAxis = 1;
	else
		splitAxis = 2;

	/* QUICKSORT-like partitioning iterations until the entry referenced by 
	   'pivot' imposes an ordering wrt. all other observations in the range */
	quickPartition(sortStart, sortEnd, pivot, splitAxis);
	short splitPos = (*pivot)->pos[splitAxis];

	/* Update the heap permutation and record the splitting axis */
	heapPermutation[heapIndex] = *pivot - *basePtr;
	(*pivot)->splitAxis = splitAxis;

	if (pivot > sortStart) {
		if (pivot > sortStart + 1) {
			/* There are more then two elements on the left
			   subtree. Balance them recursively */
			std::swap(aabb.max[splitAxis], splitPos);
			balanceRecursive(basePtr, sortStart, pivot, heapPermutation,
				aabb, leftChild(heapIndex));
			std::swap(aabb.max[splitAxis], splitPos);
		} else {
			/* Leaf node - just copy. */
			heapPermutation[leftChild(heapIndex)] = *sortStart - *basePtr;
		}
	}

	if (pivot < sortEnd - 1) {
		if (pivot < sortEnd - 2) {
			/* There are more then two elements on the right
			   subtree. Balance them recursively */
			std::swap(aabb.min[splitAxis], splitPos);
			balanceRecursive(basePtr, pivot+1, sortEnd, heapPermutation, 
				aabb, rightChild(heapIndex));
			std::swap(aabb.min[splitAxis], splitPos);
		} else {
			/* Leaf node - just copy. */
			heapPermutation[rightChild(heapIndex)] = *(sortEnd-1) - *basePtr;
		}
	}
}
	
void KDTree::obbQuery(const OBB &obb, std::vector<Observation *> &results) {
	size_t stack[MAX_KDTREE_DEPTH];
	int index = 1, stackPos = 1;

	stack[0] = 0;

	while (index > 0) {
		Observation *observation = &m_observations[index];

		if (isInnerNode(index)) {
			OBB::EClassification classify = 
				obb.classify(observation->getSplitAxis(), observation->getSplitPos()*m_scale[observation->getSplitAxis()]);

			if (classify == OBB::ELeft) {
				index = leftChild(index);
			} else if (classify == OBB::ERight) {
				if (hasRightChild(index))
					index = rightChild(index);
				else
					index = stack[--stackPos];
			} else {
				if (hasRightChild(index))
					stack[stackPos++] = rightChild(index);
				index = leftChild(index);
			}
		} else {
			 index = stack[--stackPos];
		}

		Point3 pos = observation->getPosition();
		pos.x *= m_scale.x; pos.y *= m_scale.y; pos.z *= m_scale.z;

		if (!observation->isUsed() && obb.contains(pos))
			results.push_back(observation);
	}
}

std::string KDTree::toString() const {
	std::ostringstream oss;
	oss << "KDTree[" << endl
		<< "  aabb = " << m_aabb.toString() << "," << endl
		<< "  observationCount = " << m_observationCount << "," << endl
		<< "  balanced = " << m_balanced << endl
		<< "]";
	return oss.str();
}

