#include <hair/obb.h>
#include <GL/gl.h>

bool OBB::overlaps(const OBB &other) const {
	Vector3 D = m_center - other.getCenter();

	/* Separating axis test for local axes */
	for (int i=0; i<3; ++i) {
		Vector3 axis = m_axes[i];
		double b0 = absDot(other.getAxis(0), axis) * other.getExtents(0);
		double b1 = absDot(other.getAxis(1), axis) * other.getExtents(1);
		double b2 = absDot(other.getAxis(2), axis) * other.getExtents(2);

		if (m_extents[i]+b0+b1+b2 < absDot(D, axis))
			return false;
	}

	/* Separating axis test for axes from the other OBB */
	for (int i=0; i<3; ++i) {
		Vector3 axis = other.getAxis(i);
		double a0 = absDot(m_axes[0], axis) * m_extents[0];
		double a1 = absDot(m_axes[1], axis) * m_extents[1];
		double a2 = absDot(m_axes[2], axis) * m_extents[2];

		if (a0+a1+a2+other.getExtents(i) < absDot(D, axis))
			return false;
	}

	/* Test cross product axes */
	for (int i=0; i<3; ++i) {
		for (int j=0; j<3; ++j) {
			Vector3 axis = cross(m_axes[i], other.getAxis(j));
			double a0 = absDot(m_axes[0], axis) * m_extents[0];
			double a1 = absDot(m_axes[1], axis) * m_extents[1];
			double a2 = absDot(m_axes[2], axis) * m_extents[2];
			double b0 = absDot(other.getAxis(0), axis) * other.getExtents(0);
			double b1 = absDot(other.getAxis(1), axis) * other.getExtents(1);
			double b2 = absDot(other.getAxis(2), axis) * other.getExtents(2);

			if (a0+a1+a2+b0+b1+b2 < absDot(D, axis))
				return false;
		}
	}

	return true;
}
	
OBB OBB::transform(float A[4][4]) const {
	OBB obb;

	/* Upper left 3x3 part */
	for (int i=0; i<3; ++i) {
		for (int j=0; j<3; ++j) {
			obb.m_center[i] += A[i][j] * m_center[j];

			for (int k=0; k<3; ++k)
				obb.m_axes[k][i] += A[i][j] * m_axes[k][j];
		}
	}

	/* Affine part */
	obb.m_center[0] += A[0][3];
	obb.m_center[1] += A[1][3];
	obb.m_center[2] += A[2][3];
	obb.m_extents = m_extents;

	/* Normalize */
	for (int i=0; i<3; ++i) {
		Float length = obb.m_axes[i].length();
		obb.m_extents[i] *= length;
		obb.m_axes[i] /= length;
	}

	return obb;
}

Vector3 OBB::getWorldExtents() const {
	Vector3 extents;

	for (int i=0; i<3; ++i)
		for (int j=0; j<3; ++j)
			extents[j] += std::abs(m_axes[i][j] * m_extents[i]);

	return extents;
}

bool OBB::contains(const Point3 &p) const {
	Vector3 rel = p - m_center;

	for (int i=0; i<3; ++i) {
		if (absDot(rel, m_axes[i]) > m_extents[i])
			return false;
	}
	return true;
}

Point3 OBB::findClosest(const Point3 &p) const {
	Vector3 rel = p - m_center;
	Point3 result = m_center;

	for (int i=0; i<3; ++i) {
		Float proj = dot(rel, m_axes[i]);
		Float clamped = std::max(std::min(proj, m_extents[i]), 
			-m_extents[i]);
		result += m_axes[i] * clamped;
	}
	return result;
}

void OBB::drawGL() const {
	const Vector3 v0 = m_axes[0] * m_extents[0];
	const Vector3 v1 = m_axes[1] * m_extents[1];
	const Vector3 v2 = m_axes[2] * m_extents[2];

	#define V(a,b,c) glVertex3f( \
		m_center.x a v0.x b v1.x c v2.x, \
		m_center.y a v0.y b v1.y c v2.y, \
		m_center.z a v0.z b v1.z c v2.z)
		
		glBegin(GL_LINE_LOOP); V(+,-,+); V(+,-,-); V(+,+,-); V(+,+,+); glEnd();
		glBegin(GL_LINE_LOOP); V(+,+,+); V(+,+,-); V(-,+,-); V(-,+,+); glEnd();
		glBegin(GL_LINE_LOOP); V(+,+,+); V(-,+,+); V(-,-,+); V(+,-,+); glEnd();
		glBegin(GL_LINE_LOOP); V(-,-,+); V(-,+,+); V(-,+,-); V(-,-,-); glEnd();
		glBegin(GL_LINE_LOOP); V(-,-,+); V(-,-,-); V(+,-,-); V(+,-,+); glEnd();
		glBegin(GL_LINE_LOOP); V(-,-,-); V(-,+,-); V(+,+,-); V(+,-,-); glEnd();
	#undef V
}

Point3 OBB::getCorner(int corner) const {
	const Vector3 v0 = m_axes[0] * m_extents[0];
	const Vector3 v1 = m_axes[1] * m_extents[1];
	const Vector3 v2 = m_axes[2] * m_extents[2];

	return m_center
		+ v0 * ((corner & 1) ? -1.0 : 1.0)
		+ v1 * ((corner & 2) ? -1.0 : 1.0)
		+ v2 * ((corner & 4) ? -1.0 : 1.0);
}

OBB::EClassification OBB::classify(int axis, Float value) const {
	Float axis0 = std::abs(m_axes[0][axis]) * m_extents[0];
	Float axis1 = std::abs(m_axes[1][axis]) * m_extents[1];
	Float axis2 = std::abs(m_axes[2][axis]) * m_extents[2];
	Float left  = m_center[axis] - axis0 - axis1 - axis2;
	Float right = m_center[axis] + axis0 + axis1 + axis2;

	if (left <= value && right <= value)
		return ELeft;
	else if (left >= value && right >= value)
		return ERight;
	else
		return EBoth;
}

std::string OBB::toString() const {
	std::ostringstream oss;
	oss << "OBB[" << endl
		<< "  center = " << m_center.toString() << "," << endl
		<< "  axes[0] = " << m_axes[0].toString() << "," << endl
		<< "  axes[1] = " << m_axes[1].toString() << "," << endl
		<< "  axes[2] = " << m_axes[2].toString() << "," << endl
		<< "  extents = " << m_extents.toString() << endl
		<< "]";
	return oss.str();
}

