#include <hair/parameters.h>
#include <boost/program_options.hpp>
#include <boost/tokenizer.hpp>
#include <boost/foreach.hpp>
#include <fstream>

namespace po = boost::program_options;

Parameters::Parameters(int argc, char **argv) : cosTheta(NULL), sinTheta(NULL), filterOrientations(NULL), ridgeOrientations(NULL) {
	std::string confFile;
	std::string sheetSizeStr;

	po::options_description generic("Generic options"),
		config("Configuration options");
	generic.add_options()
		("version,v", "Print the version string")
		("help", "Print a description of all parameters")
		("filter", "Image filtering mode")
		("file,f", po::value<std::string>(&confFile)->default_value("hair.conf"),
			"Load configuration options from an external file");
	config.add_options()
		("input", po::value<std::string>(&input)->default_value("data-%i.png"),
			"Input filename format (16 bit, 1 channel)")
		("filterOutput", po::value<std::string>(&filterOutput)->default_value("filtered-%i.pix"),
			"Output filename format for filtered data")
		("numAngles", po::value<int>(&numAngles)->default_value(12),
			"Total number of angles in the capture")
		("numDepths", po::value<int>(&numDepths)->default_value(400),
			"World-space offsets of the captured focus depths and associated filename prefixes")
		("translateStart", po::value<float>(&translateStart)->default_value(-50),
			"Initial translation offset in world-space units")
		("translateEnd", po::value<float>(&translateEnd)->default_value(50),
			"Final translation offset in world-space  units")
		("xfov", po::value<float>(&xfov)->default_value(11.0f),
			"Field of view in degrees")
		("fPlaneDist", po::value<float>(&fPlaneDist)->default_value(278.0f),
			"Distance to the focus plane")
		("xres", po::value<int>(&xres)->default_value(768),
			"Per-color X resolution on the camera sensor")
		("yres", po::value<int>(&yres)->default_value(768),
			"Per-color Y resolution on the camera sensor")
		("gaborCount", po::value<int>(&gaborCount)->default_value(32),
			"Number of gabor filters used to estimate orientation")
		("hysteresisLow", po::value<int>(&hysteresisLow)->default_value(30),
			"Lower bound for hysteresis thresholding")
		("hysteresisHigh", po::value<int>(&hysteresisHigh)->default_value(100),
			"Upper bound for hysteresis thresholding")
		("gaussianExtents", po::value<int>(&gaussianExtents)->default_value(1),
			"Extents of the gaussian pre-filtering pass")
		("centerX", po::value<float>(&centerX)->default_value(0),
			"Projected X coordinate of the rotation axis")
		("filterOffset", po::value<float>(&filterOffset)->default_value(-100),
			"Filter offset (adjusts the amount of noise in dark regions)")
		("correctionCutoff", po::value<float>(&correctionCutoff)->default_value(.5f),
			"Growing algorithm: max. tangential correction offset in pixels")
		("directionCutoff", po::value<float>(&directionCutoff)->default_value(0.75f),
			"Growing algorithm: maximum angle to the previous fitted direction (in degrees)")
		("threadCount", po::value<int>(&threadCount)->default_value(1),
			"Thread count during reconstruction")
		("minObservations", po::value<int>(&minObservations)->default_value(2),
			"Required number of observations for the direction/correction fitting algorithms")
		("sheetSize", po::value<std::string>(&sheetSizeStr)->default_value("80, 30, 4"),
			"Query sheet size");

	/* Parse generic option first */
	po::options_description cmdline_options;
	cmdline_options.add(generic).add(config);
	po::variables_map vmFirstTry, vm;
	po::store(po::parse_command_line(argc, argv, cmdline_options), vmFirstTry);
	po::notify(vmFirstTry);

	if (vmFirstTry.count("help")) {
		cout << cmdline_options << "\n";
		exit(0);
	} else if (vmFirstTry.count("version")) {
		cout << VERSION_STRING << "\n";
		exit(0);
	}

	if (vmFirstTry.count("filter")) 
		filter = true;
	else
		filter = false;

	std::ifstream ifs(confFile.c_str());
	if (ifs.fail()) {
		cerr << "Error - could not open '" << confFile << "'!" << endl;
		exit(-1);
	}

	/* Now that we know the name of the configuration file, do it once more. */
	po::options_description configfile_options;
	configfile_options.add(config);
	po::store(po::parse_command_line(argc, argv, cmdline_options), vm);
    po::store(po::parse_config_file(ifs, configfile_options), vm);
	po::notify(vm);

	boost::char_separator<char> sep1(" ,");
	boost::tokenizer<boost::char_separator<char> > tok(sheetSizeStr, sep1);
	int pos = 0;
	BOOST_FOREACH(std::string s, tok) {
		if (pos < 3) {
			sheetSize[pos++] = boost::lexical_cast<Float>(s); 
		} else if (pos == 3) {
			angleThreshold = boost::lexical_cast<Float>(s); 
			pos++;
		}
	}
	assert(pos == 4);
	directionCutoff *= M_PI/180.0;
	precompute();
}

void Parameters::precompute() {
	translateStep = (translateEnd-translateStart) / (float) (numDepths-1); 
	absTranslateStep = std::abs(translateStep);
	angleStep = (float) ((2*M_PI) / numAngles);

	pixelWidth = 2.0f * fPlaneDist 
		* std::tan(xfov*M_PI/180.0f*.5f) / (float) xres;
	centerDepth = (int) (-translateStart/translateStep);
	aspectZ = pixelWidth / translateStep;

	float tanX = std::tan(degToRad(xfov) / 2);
	f = .5f * xres / tanX;

	cosTheta = new float[numAngles];
	sinTheta = new float[numAngles];
	filterOrientations = new float[gaborCount];
	ridgeOrientations = new float[gaborCount];

	for (int i=0; i<gaborCount; ++i) {
		filterOrientations[i] = (float) M_PI * ((float)i / gaborCount);
		ridgeOrientations[i] = filterOrientations[i] + M_PI/2.0f;
	}

	for (int angle=0; angle<numAngles; ++angle) {
		Float angleRad = (2*M_PI/numAngles)*angle;
		cosTheta[angle] = std::cos(angleRad);
		sinTheta[angle] = std::sin(angleRad);
	}
}

Parameters::~Parameters() {
	if (cosTheta)
		delete[] cosTheta;
	if (sinTheta)
		delete[] sinTheta;
	if (filterOrientations)
		delete[] filterOrientations;
	if (ridgeOrientations)
		delete[] ridgeOrientations;
}

std::string Parameters::toString() const {
	std::ostringstream oss;
	oss << "========================================" << endl;
	oss << "|            Parameter setup           |" << endl;
	oss << "========================================" << endl;
	oss << "  Input images 16 bit : " << input << endl;
	oss << "  Filtered output     : " << filterOutput << endl;
	oss << "  Number of angles    : " << numAngles << endl;
	oss << "  Number of depths    : " << numDepths << endl;
	oss << "  # of Gabor filters  : " << gaborCount << endl;
	oss << "  Hysteresis low      : " << hysteresisLow << endl;
	oss << "  Hysteresis high     : " << hysteresisHigh << endl;
	oss << "  Gaussian filter     : " << gaussianExtents << endl;
	oss << "  Horizontal FOV      : " << xfov << endl;
	oss << "  Focus plane dist.   : " << fPlaneDist << endl;
	oss << "  Sensor res. (X)     : " << xres << endl;
	oss << "  Sensor res. (Y)     : " << yres << endl;
	oss << "  Translation start   : " << translateStart << endl;
	oss << "  Translation end     : " << translateEnd << endl;
	oss << "  X Image Center      : " << centerX << endl;
	oss << "  Focal length        : " << f << endl;
	oss << "  Correction cutoff   : " << correctionCutoff << "px" << endl;
	oss << "  Direction cutoff    : " << directionCutoff * 180/M_PI << "°"<< endl;
	oss << "  Pixel width         : " << pixelWidth << "mm"<< endl;
	oss << "  Thread count        : " << threadCount << endl;
	oss << "  Oriented box size   : " << sheetSize.toString() << endl;
	oss << "  Angular tolerance   : " << angleThreshold << "°" << endl;
	oss << "  Min. observations   : " << minObservations << endl;
	oss << "========================================" << endl;
	return oss.str();
}
