#include <hair/wthread.h>
	
const int gridRes = 16;			    /// Coarse 16x16x16 grid
const Float cellWidth = 16;		    /// Each cell covers 16^3 pixels
const int minObservations = 5;	    /// Minimum # of observations per cell
const int minObservationAngles = 4; /// Minimum # of observation angles
const Float sideLength = gridRes * cellWidth;

WorkerThread::WorkerThread(const AABB &aabb, DataModel *dataModel)
		: m_dataModel(dataModel), m_growing(false) {
	const int numCells = gridRes*gridRes*gridRes*dataModel->getParameters().numAngles;
	m_cells = new uint16_t[numCells];
	m_regenerateCells = true;
	m_side = 0;
	m_aabb[0] = aabb;
	m_aabb[1] = aabb;
	Float ycenter = .5f * (aabb.max.y + aabb.min.y);
	m_aabb[0].min.y = ycenter;
	m_aabb[1].max.y = ycenter;
	m_active = false;
}

WorkerThread::~WorkerThread() {
	delete[] m_cells;
}

void WorkerThread::run() {
	setPriority(QThread::LowestPriority);
	m_active = true;

	while (m_active && grow(false))
		;

	m_active = false;
}

/* Perform one growth step. The paramenter is set to true 
   when the 'Grow feature' button was pressed */
bool WorkerThread::grow(bool onlyOneStep) {
	EReturnCode status;
	bool success;
	const Parameters &params = m_dataModel->getParameters();

	if (!m_growing) {
		/* Search for a new seed point */
		do {
			status = findSeed(m_activeSet[0]);

			if (onlyOneStep && status == Failure)
				return false;
			else if (!onlyOneStep && !m_active)
				return false;
		} while (status == Failure);

		m_growing = true;
		if (status == Done)
			return false;

		/* Generate two seeds in opposite directions */
		m_activeSet[1] = m_activeSet[0];
		m_activeSet[1].d *= -1;
		m_activeSet[1].dOriginal *= -1;
	}

	if (onlyOneStep) {
		success = false;
		success |= (m_dataModel->grow(m_activeSet[0]) == Success);
		success |= (m_dataModel->grow(m_activeSet[1]) == Success);
		m_dataModel->refreshViews();
		if (success)
			return false;
	} else {
		do {
			success = false;
			success |= (m_dataModel->grow(m_activeSet[0]) == Success);
			success |= (m_dataModel->grow(m_activeSet[1]) == Success);
			if (!m_active)
				return false;
		} while (success);
	}

	m_growing = false;
	HairFiber fiber;
	std::vector<Point3> &p0 = m_activeSet[0].fiber.vertices;
	std::vector<Point3> &p1 = m_activeSet[1].fiber.vertices;
	size_t nPoints = p0.size() + p1.size() - 2*MAX_BLIND_STEPS;

	if (nPoints > 100) { // discard short hairs
		for (int i=(int) p0.size()-MAX_BLIND_STEPS-1; i>=0; --i)
			fiber.vertices.push_back(p0[i]);
		for (size_t i=0; i<p1.size()-MAX_BLIND_STEPS; ++i)
			fiber.vertices.push_back(p1[i]);

		/* Avoid future re-detection */
		for (int side=0; side<2; ++side) {
			for (int angle=0; angle<params.numAngles; ++angle) {
				std::set<KDTree::Observation *> &observations = m_activeSet[side].observations[angle];
				for (std::set<KDTree::Observation *>::iterator it = observations.begin(); 
						it != observations.end(); ++it) {
					(*it)->setUsed(true);
				}
			}
		}

		m_regenerateCells = true;
		m_dataModel->addFiber(fiber);
	}

	m_activeSet[0].lock();
	m_activeSet[1].lock();
	for (int side=0; side<2; ++side) {
		for (int angle=0; angle<params.numAngles; ++angle) 
			m_activeSet[side].observations[angle].clear();
	}
	p0.clear();
	p1.clear();
	m_activeSet[0].unlock();
	m_activeSet[1].unlock();

	return true;
}

/* Search for seed point used to start growing a fiber */
WorkerThread::EReturnCode WorkerThread::findSeed(ActiveSet &activeSet) {
	const Parameters &params = m_dataModel->getParameters();
	const int numCells = gridRes*gridRes*gridRes*params.numAngles;

	if (!m_seedAABB.isValid()) {
		m_seedAABB.min = m_aabb[m_side].min;
		m_seedAABB.max = m_seedAABB.min + Vector3(1,1,1)*sideLength;
	}
					
	int bestObservationCount = -1;
	int bestPixelCount = 0;
	Point3s bestPos;
	Point3 bestPosWorld;

	if (m_regenerateCells) {
		/* Determine the number of observations in each cell */
		memset(m_cells, 0, sizeof(uint16_t)*numCells);
		for (int z=0; z<gridRes; ++z) {
			for (int y=0; y<gridRes; ++y) {
				for (int x=0; x<gridRes; ++x) {
					AABB cell(
						m_seedAABB.min + Vector3(x,y,z)*cellWidth,
						m_seedAABB.min + Vector3(x+1,y+1,z+1)*cellWidth
					);
					for (int angle=0; angle<params.numAngles; ++angle) {
						m_result.clear();
						m_dataModel->obbQuery(OBB(cell), angle, m_result);
						m_cells[x + gridRes * (y + gridRes * (z + angle * gridRes))] = m_result.size();
					}
				}
			}
		}
		m_regenerateCells = false;
	}

	/* Search for the cell with the most observations */
	for (int z=0; z<gridRes; ++z) {
		for (int y=0; y<gridRes; ++y) {
			for (int x=0; x<gridRes; ++x) {
				int observationCount = 0, pixelCount = 0;

				for (int angle=0; angle<params.numAngles; ++angle) {
					int nPixels = m_cells[x + gridRes * (y + gridRes * (z + angle * gridRes))];

					if (nPixels > minObservations) {
						observationCount++;
						pixelCount += nPixels;
					}
				}

				if (observationCount > bestObservationCount ||
				(observationCount == bestObservationCount &&
					pixelCount > bestPixelCount)) {
					bestPosWorld = m_seedAABB.min + 
						(Vector3(bestPos.x + .5f, bestPos.y + .5f, bestPos.z + .5)) * cellWidth;
					if (m_previousSeeds.find(Point3s(x,y,z)) == m_previousSeeds.end()
						&& m_aabb[m_side].contains(bestPosWorld)) { 
						bestObservationCount = observationCount;
						bestPixelCount = pixelCount;
						bestPos = Point3s(x,y,z);
					}
				}
			}
		}
	}

	/* If the found number of observations was not satisfactory, 
	   keep traversing the search volume */
	if (bestObservationCount < minObservationAngles) {
		m_seedAABB.min.x += sideLength;
		m_seedAABB.max.x += sideLength;
		m_previousSeeds.clear();
		m_regenerateCells = true;

		if (m_seedAABB.min.x > m_aabb[m_side].max.x) {
			m_seedAABB.min.x = m_aabb[m_side].min.x;
			m_seedAABB.max.x = m_aabb[m_side].min.x + sideLength;
			m_seedAABB.min.z += sideLength;
			m_seedAABB.max.z += sideLength;

			if (m_seedAABB.min.z > m_aabb[m_side].max.z) {
				m_seedAABB.min.z = m_aabb[m_side].min.z;
				m_seedAABB.max.z = m_aabb[m_side].min.z + sideLength;

				m_seedAABB.min.y += sideLength;
				m_seedAABB.max.y += sideLength;
			
				if (m_seedAABB.min.y > m_aabb[m_side].max.y) {
					if (++m_side>=2)
						return Done; /* Done! */
					m_seedAABB = AABB();
				}
			}
		}
	
		m_dataModel->refreshViews();
		return Failure;
	}

	activeSet.p = bestPosWorld;
	activeSet.nSteps = activeSet.blindSteps = activeSet.nActive = 0;
	activeSet.fiber = HairFiber();

	for (int angle=0; angle<params.numAngles; ++angle) {
		activeSet.fiber.seen[angle] = false;
		activeSet.active[angle] = false;
	}

	/* Look for observations near the selected position and compute averaged 2D orientations */
	OBB searchRegion(AABB(
		bestPosWorld - Vector3(1,1,1)*cellWidth,
		bestPosWorld + Vector3(1,1,1)*cellWidth
	));

	activeSet.lock();
	for (int angle=0; angle<params.numAngles; ++angle) {
		m_result.clear();
		m_dataModel->obbQuery(searchRegion, angle, m_result);

		if (m_result.size() > 0) {
			Vector2 direction;
			for (size_t j=0; j<m_result.size(); ++j) {
				Float orientation = params.ridgeOrientations[m_result[j]->getOrientationIndex()];
				Vector2 d(std::cos(orientation), std::sin(orientation));
				if (dot(d, direction) < 0)
					d = -d;
				direction += d;
			}
			activeSet.orientation[angle] = std::atan(direction.y/direction.x);
			activeSet.active[angle] = true;
			activeSet.nActive++;
		}
	}
	m_previousSeeds.insert(bestPos);
	activeSet.unlock();

	/* Use the observed 2D orientations to fit a 3D fiber direction */
	if (!m_dataModel->solveDirection(activeSet, activeSet.d))
		return Failure;
	
	activeSet.dOriginal = activeSet.d;
	
	/* Snap onto the fiber */
	Vector3 correction;
	if (!m_dataModel->solveCorrection(activeSet, activeSet.p, correction))
		return Failure;

	activeSet.p += correction;
	activeSet.pOriginal = activeSet.p;

	/* Try to reconstruct hairs with a coherent overall direction 
	   This is important for the rendering side, which keeps track of 
	   scale orientations on the surface of a hair.
	*/
	if (activeSet.d.y < 0)
		activeSet.d *= -1;

	return Success;
}
